
from scipy.linalg import \
    inv
from traits.api import \
    Array, Float, \
    Instance, Int

from ibvpy.fets import FETSEval
from ibvpy.tmodel.mats_eval import MATSEval
import numpy as np


#-------------------------------------------------------------------------
# FETS2D4Q8U - 8 nodes subparametric quadrilateral (2D, quadratic, serendipity family)
#-------------------------------------------------------------------------
#-------------------------------------------------------------------------
# Element Information:
#-------------------------------------------------------------------------
#
# The order of the field approximation is higher then the order of the geometry
# approximation (subparametric element).
# The implemented shape functions are derived (in femple) based
# on the following ordering of the nodes of the parent element.
#
#            _node_coord_map_dof = np.array( Float, (8,2),
#                                 [[ -1.,-1. ],
#                                  [  1.,-1. ],
#                                  [  1., 1. ],
#                                  [ -1., 1. ],
#                                  [  0.,-1. ],
#                                  [  1., 0. ],
#                                  [  0., 1. ],
#                                  [ -1., 0. ]])
#
# The ordering of the nodes of the parent element used for the geometry approximation
# is defined in '_node_coord_map_geo' (see code below)
# and the (linear) shape functions are derived by formula
#
#-------------------------------------------------------------------------
class FETS2D4Q8U(FETSEval):
    debug_on = True

    mats_eval = Instance(MATSEval)

    # Dimensional mapping
    dim_slice = slice(0, 2)

    n_e_dofs = Int(2 * 8)
    t = Float(1.0, label='thickness')

    # Integration parameters
    #
    ngp_r = 3
    ngp_s = 3

    dof_r = Array(value=[[-1., -1.],
                         [1., -1.],
                         [1., 1.],
                         [-1., 1.],
                         [0., -1.],
                         [1., 0.],
                         [0., 1.],
                         [-1., 0.]])
    geo_r = Array(value=[[-1, -1], [1, -1], [1, 1], [-1, 1]])
    #
    vtk_r = Array(value=[[-1., -1.],
                         [0., -1.],
                         [1., -1.],
                         [-1., 0.],
                         [1., 0.],
                         [-1., 1.],
                         [0., 1.],
                         [1., 1.]])
    vtk_cells = [[0, 2, 7, 5, 1, 4, 6, 3]]
    vtk_cell_types = 'QuadraticQuad'

    n_nodal_dofs = 2

    # Ordering of the nodes of the parent element used for the geometry
    # approximation
    _node_coord_map_geo = Array(Float, (4, 2),
                                [[-1., -1.],
                                 [1., -1.],
                                 [1., 1.],
                                 [-1., 1.]])

    #---------------------------------------------------------------------
    # Method required to represent the element geometry
    #---------------------------------------------------------------------
    def get_N_geo_mtx(self, r_pnt):
        '''
        Return the value of shape functions for the specified local coordinate r_pnt
        '''
        cx = self._node_coord_map_geo
        N_geo_mtx = np.array(
            [[1 / 4. * (1 + r_pnt[0] * cx[i, 0]) * (1 + r_pnt[1] * cx[i, 1]) for i in range(0, 4)]])
        return N_geo_mtx

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives.
        Used for the conrcution of the Jacobi matrix.

        @TODO - the B matrix is used
        just for uniaxial bar here with a trivial differential
        operator.
        '''
        cx = self._node_coord_map_geo
        dNr_geo_mtx = np.array([[1 / 4. * cx[i, 0] * (1 + r_pnt[1] * cx[i, 1]) for i in range(0, 4)],
                                [1 / 4. * cx[i, 1] * (1 + r_pnt[0] * cx[i, 0]) for i in range(0, 4)]])
        return dNr_geo_mtx

    #-------------------------------------------------------------------------
    # Method delivering the shape functions for the field variables and their derivatives
    #-------------------------------------------------------------------------
    def get_N_mtx(self, r_pnt):
        '''
        Returns the matrix of the shape functions (derived in femple) used for the field 
        approximation containing zero entries. The number of rows corresponds to the number 
        of nodal dofs. The matrix is evaluated for the specified local coordinate r_pnt.
        '''
        N_dof = np.zeros((1, 8))
        N_dof[0, 0] = -((-1 + r_pnt[1]) * (-1 + r_pnt[0])
                        * (r_pnt[0] + 1 + r_pnt[1])) / 4.0
        N_dof[0, 1] = -((-1 + r_pnt[1]) * (1 + r_pnt[0])
                        * (r_pnt[0] - 1 - r_pnt[1])) / 4.0
        N_dof[0, 2] = (
            (1 + r_pnt[1]) * (1 + r_pnt[0]) * (r_pnt[0] - 1 + r_pnt[1])) / 4.0
        N_dof[0, 3] = (
            (1 + r_pnt[1]) * (-1 + r_pnt[0]) * (r_pnt[0] + 1 - r_pnt[1])) / 4.0
        N_dof[0, 4] = (
            (-1 + r_pnt[0]) * (1 + r_pnt[0]) * (-1 + r_pnt[1])) / 2.0
        N_dof[0, 5] = - \
            ((-1 + r_pnt[1]) * (1 + r_pnt[1]) * (1 + r_pnt[0])) / 2.0
        N_dof[0, 6] = - \
            ((-1 + r_pnt[0]) * (1 + r_pnt[0]) * (1 + r_pnt[1])) / 2.0
        N_dof[0, 7] = (
            (-1 + r_pnt[1]) * (1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0

        I_mtx = np.identity(self.n_nodal_dofs, float)
        N_mtx_list = [I_mtx * N_dof[0, i] for i in range(0, N_dof.shape[1])]
        N_mtx = np.hstack(N_mtx_list)
        return N_mtx

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions (derived in femple) 
        used for the field approximation
        '''
        dNr_mtx = np.zeros((2, 8), dtype='float64')
        dNr_mtx[0, 0] = -((-1 + r_pnt[1]) * (r_pnt[0] + 1 + r_pnt[1])
                          ) / 4.0 - ((-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 4.0
        dNr_mtx[0, 1] = -((-1 + r_pnt[1]) * (r_pnt[0] - 1 - r_pnt[1])
                          ) / 4.0 - ((-1 + r_pnt[1]) * (1 + r_pnt[0])) / 4.0
        dNr_mtx[0, 2] = ((1 + r_pnt[1]) * (r_pnt[0] - 1 + r_pnt[1])
                         ) / 4.0 + ((1 + r_pnt[1]) * (1 + r_pnt[0])) / 4.0
        dNr_mtx[0, 3] = ((1 + r_pnt[1]) * (r_pnt[0] + 1 - r_pnt[1])
                         ) / 4.0 + ((1 + r_pnt[1]) * (-1 + r_pnt[0])) / 4.0
        dNr_mtx[0, 4] = ((-1 + r_pnt[1]) * (1 + r_pnt[0])) / \
            2.0 + ((-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0
        dNr_mtx[0, 5] = -((-1 + r_pnt[1]) * (1 + r_pnt[1])) / 2.0
        dNr_mtx[0, 6] = -((1 + r_pnt[1]) * (1 + r_pnt[0])) / \
            2.0 - ((1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0
        dNr_mtx[0, 7] = ((-1 + r_pnt[1]) * (1 + r_pnt[1])) / 2.0
        dNr_mtx[1, 0] = -((-1 + r_pnt[0]) * (r_pnt[0] + 1 + r_pnt[1])
                          ) / 4.0 - ((-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 4.0
        dNr_mtx[1, 1] = -((1 + r_pnt[0]) * (r_pnt[0] - 1 - r_pnt[1])
                          ) / 4.0 + ((-1 + r_pnt[1]) * (1 + r_pnt[0])) / 4.0
        dNr_mtx[1, 2] = ((1 + r_pnt[0]) * (r_pnt[0] - 1 + r_pnt[1])
                         ) / 4.0 + ((1 + r_pnt[1]) * (1 + r_pnt[0])) / 4.0
        dNr_mtx[1, 3] = ((-1 + r_pnt[0]) * (r_pnt[0] + 1 - r_pnt[1])
                         ) / 4.0 - ((1 + r_pnt[1]) * (-1 + r_pnt[0])) / 4.0
        dNr_mtx[1, 4] = ((-1 + r_pnt[0]) * (1 + r_pnt[0])) / 2.0
        dNr_mtx[1, 5] = -((1 + r_pnt[1]) * (1 + r_pnt[0])) / \
            2.0 - ((-1 + r_pnt[1]) * (1 + r_pnt[0])) / 2.0
        dNr_mtx[1, 6] = -((-1 + r_pnt[0]) * (1 + r_pnt[0])) / 2.0
        dNr_mtx[1, 7] = ((1 + r_pnt[1]) * (-1 + r_pnt[0])) / \
            2.0 + ((-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 2.0
        return dNr_mtx

    def get_B_mtx(self, r_pnt, X_mtx):
        J_mtx = self.get_J_mtx(r_pnt, X_mtx)
        dNr_mtx = self.get_dNr_mtx(r_pnt)
        dNx_mtx = np.dot(inv(J_mtx), dNr_mtx)
        Bx_mtx = np.zeros((3, 16), dtype='float64')
        for i in range(0, 8):
            Bx_mtx[0, i * 2] = dNx_mtx[0, i]
            Bx_mtx[1, i * 2 + 1] = dNx_mtx[1, i]
            Bx_mtx[2, i * 2] = dNx_mtx[1, i]
            Bx_mtx[2, i * 2 + 1] = dNx_mtx[0, i]
        return Bx_mtx


#----------------------- example --------------------

if __name__ == '__main__':
    from ibvpy.api import \
        TStepper as TS, RTraceDomainListField, TLoop, \
        TLine, BCDofGroup

    #from ibvpy.tmodel.mats2D.mats_cmdm2D.mats_cmdm2D import MATS2DMicroplaneDamage
    #from ibvpy.tmodel.mats2D.mats2D_sdamage.mats2D_sdamage import MATS2DScalarDamage
    #from ibvpy.tmodel.mats2D.mats2D_sdamage.strain_norm2d import *
    from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
#    fets_eval = FETS2D4Q8U(mats_eval = MACMDM())
    #fets_eval = FETS2D4Q8U(mats_eval = MATS2DScalarDamage(strain_norm_type = 'Euclidean'))
    fets_eval = FETS2D4Q8U(mats_eval=MATS2DElastic())

    # Define a mesh domain adaptor as a cached property to
    # be constracted on demand

    from ibvpy.mesh.fe_grid import FEGrid

    # Discretization
    domain = FEGrid(coord_max=(3., 3., 0.),
                    shape=(1, 1),
                    fets_eval=fets_eval)

    # Put the tseval (time-stepper) into the spatial context of the
    # discretization and specify the response tracers to evaluate there.
    right_dof = 2
    ts = TS(
        sdomain=domain,
        # conversion to list (square brackets) is only necessary for slicing of
        # single dofs, e.g "get_left_dofs()[0,1]"

        # Boundary conditions for three-point-bendig:
        #         bcond_list =  [ BCDof(var='u', dof = i, value = 0.) for i in [domain.get_bottom_left_dofs()[0,0]]  ] +
        #                    [ BCDof(var='u', dof = i, value = 0.) for i in [domain.get_bottom_left_dofs()[0,1]] ] +
        #                    [ BCDof(var='u', dof = i, value = 0.) for i in [domain.get_bottom_right_dofs()[0,1]] ] +
        #                    [ BCDof(var='u', dof = i, value = 0.002 ) for i in [domain.get_top_middle_dofs()[0,1]] ],

        # Boundary conditions for shear force applied at the right border:
        bcond_list=[BCDofGroup(var='u', value=0., dims=[0, 1],
                               get_dof_method=domain.get_left_dofs),
                    #                         BCDofGroup( var='u', value = 0., dims = [1],
                    # get_dof_method = domain.get_bottom_right_dofs ),
                    BCDofGroup(var='u', value=0.002, dims=[1],
                               get_dof_method=domain.get_right_dofs)],


        rtrace_list=[
            #                         RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
            #                               var_y = 'F_int', idx_y = right_dof,
            #                               var_x = 'U_k', idx_x = right_dof),
            #                         RTraceDomainListField(name = 'Stress' ,
            #                         var = 'sig_app', idx = 0,
            #                         record_on = 'update'),
            #                     RTraceDomainListField(name = 'Displacement' ,
            #                                    var = 'u', idx = 1),
            RTraceDomainListField(name='Stress',
                                  #position = 'int_pnts',
                                  var='sig_app', idx=1),
            #                             RTraceDomainListField(name = 'N0' ,
            #                                          var = 'N_mtx', idx = 0,
            #                                          record_on = 'update')
        ])

    # Add the time-loop control
    #
    tl = TLoop(tstepper=ts,
               tline=TLine(min=0.0, step=1., max=1.0))

    tl.eval()
    # Put the whole stuff into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp(ibv_resource=tl)
    app.main()



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Sep 4, 2009 by: rch

from ibvpy.util.traits.either_type import \
    EitherType
from ibvpy.tmodel.mats1D.mats1D_elastic.mats1D_elastic import MATS1DElastic
from ibvpy.tmodel.mats1D.mats1D_damage.mats1D_damage import MATS1DDamage
from ibvpy.tmodel.mats1D.mats1D_plastic.mats1D_plastic import MATS1DPlastic

from ibvpy.tmodel.matsXD.matsXD_explore import MATSXDExplore

class MATS1DExplore( MATSXDExplore ):
    '''
    Simulate the loading histories of a material point in 1D space.
    
    Specify the loading scenarios.
    '''
    mats_eval = EitherType( klasses = [ MATS1DElastic, 
                                        MATS1DDamage,
                                        MATS1DPlastic ] )
    def _mats_eval_default(self):
        return MATS1DElastic()




from numpy import array, dot
from scipy.linalg import \
    inv

from traits.api import \
    Int, Array
from ibvpy.fets.fets_eval import FETSEval


#-----------------------------------------------------------------------------
# FEBar1D
#-----------------------------------------------------------------------------
class FETS1D2L(FETSEval):
    '''
    Fe Bar 2 nodes, deformation
    '''

    debug_on = True

    # Dimensional mapping
    dim_slice = slice(0, 1)

    n_e_dofs = Int(2)
    n_nodal_dofs = Int(1)

    dof_r = Array(value=[[-1], [1]])
    geo_r = Array(value=[[-1], [1]])
    vtk_r = Array(value=[[-1.], [1.]])
    vtk_cells = [[0, 1]]
    vtk_cell_types = 'Line'

    def _get_ip_coords(self):
        offset = 1e-6
        return array([[-1 + offset, 0., 0.], [1 - offset, 0., 0.]])

    def _get_ip_weights(self):
        return array([[1.], [1.]], dtype=float)

    # Integration parameters
    #
    ngp_r = 2

    def get_N_geo_mtx(self, r_pnt):
        '''
        Return geometric shape functions
        @param r_pnt:
        '''
        r = r_pnt[0]
        N_mtx = array([[0.5 - r / 2., 0.5 + r / 2.]])
        return N_mtx

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives.
        Used for the conrcution of the Jacobi matrix.
        '''
        return array([[-1. / 2, 1. / 2]])

    def get_N_mtx(self, r_pnt):
        '''
        Return shape functions
        @param r_pnt:local coordinates
        '''
        return self.get_N_geo_mtx(r_pnt)

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions
        '''
        return self.get_dNr_geo_mtx(r_pnt)

    def get_B_mtx(self, r, X):
        '''
        Return kinematic matrix
        @param r:local coordinates
        @param X:global coordinates
        '''
        J_mtx = self.get_J_mtx(r, X)
        dNr_mtx = self.get_dNr_mtx(r)
        B_mtx = dot(inv(J_mtx), dNr_mtx)
        return B_mtx

#----------------------- example --------------------


from ibvpy.plugins.ibvpy_app import IBVPyApp


def __demo__():

    from ibvpy.api import \
        TStepper as TS, RTDofGraph, RTraceDomainListField, TLoop, \
        TLine, BCDof
    from ibvpy.tmodel.mats1D.mats1D_elastic.mats1D_elastic import MATS1DElastic

    fets_eval = FETS1D2L(mats_eval=MATS1DElastic(E=10.))
    from ibvpy.mesh.fe_grid import FEGrid

    # Discretization
    domain = FEGrid(coord_max=(3.,),
                    shape=(3,),
                    fets_eval=fets_eval)

    ts = TS(dof_resultants=True,
            sdomain=domain,
            bcond_list=[BCDof(var='u', dof=0, value=0.),
                        BCDof(var='f', dof=3, value=1,)],
            rtrace_list=[RTDofGraph(name='Fi,right over u_right (iteration)',
                                    var_y='F_int', idx_y=0,
                                    var_x='U_k', idx_x=1),
                         RTraceDomainListField(name='Stress',
                                               var='sig_app', idx=0),
                         RTraceDomainListField(name='Displacement',
                                               var='u', idx=0,
                                               warp=True),
                         RTraceDomainListField(name='N0',
                                               var='N_mtx', idx=0,
                                               record_on='update')
                         ]
            )

    # Add the time-loop control
    tloop = TLoop(tstepper=ts,
                  tline=TLine(min=0.0, step=0.5, max=1.0))

    print('---- result ----')
    print(tloop.eval())
    print(ts.F_int)
    print(ts.rtrace_list[0].trace.ydata)

    # Put the whole stuff into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    app = IBVPyApp(ibv_resource=tloop)
    app.main()


if __name__ == '__main__':
    __demo__()



'''
Created on Feb 11, 2018

@author: rch
'''

import os
from ibvpy.mathkit.tensor import EPS

# from mayavi import mlab
# from tvtk.api import tvtk
from ibvpy.view.plot3d.viz3d import Vis3D, Viz3D

import numpy as np
import traits.api as tr


class Vis3DLattice(Vis3D):
    r'''Record the response triggered by the hist object the time increments
    upon a record event.
    '''
    var = tr.Str('<unnamed>')

    extension = '.npz'

    def setup(self):
        self.new_dir()
        ts = self.tstep
        fe_domain = ts.fe_domain
        X_Ia_list = []
        I_Li_list = []
        Xm_Lia_list = []
        # Consistently with other adaptors, this one allows for
        # a loop over several lattice domains.
        for domain in fe_domain:
            xdomain = domain.xmodel
            var_function = domain.tmodel.var_dict.get(self.var, None)
            if var_function == None or xdomain.hidden:
                continue
            DELTA_x_ab = xdomain.vtk_expand_operator
            X_Ia = np.einsum(
                'Ia,ab->Ib',
                xdomain.X_Ia, DELTA_x_ab
            )
            Xm_Lia = np.einsum(
                'Lib,ba->Lia',
                xdomain.Xm_Lia, DELTA_x_ab
            )
            X_Ia_list.append(X_Ia)
            Xm_Lia_list.append(Xm_Lia)
            I_Li_list.append(xdomain.I_Li)
        if len(I_Li_list) == 0:
            raise ValueError(
                'Empty output for field variable %s in model %s' %
                (self.var, domain.tmodel)
            )
        X_Ia = np.hstack(X_Ia_list)
        I_Li = np.hstack(I_Li_list)
        fname = 'lattice_tessellation'
        target_file = os.path.join(
            self.dir, fname.replace('.', '_')
        ) + '.' + self.extension
        np.savez(target_file,
                 X_Ia=X_Ia, I_Li=I_Li, Xm_Lia=Xm_Lia)
        self.lattice_file = target_file

    def update(self):
        ts = self.tstep
        U = ts.U_k
        t = ts.t_n1
        U_Ia_list = []
        U_Lipa_list = []
        state_fields = []
        for domain in ts.fe_domain:
            xdomain = domain.xmodel
            state_field = domain.state_n.get(self.var, None)
            if (xdomain.hidden):
                # If the state variable not present in the domain, skip
                continue
            if (state_field is not None):
                state_fields.append(state_field.flatten())
            DELTA_x_ab = xdomain.vtk_expand_operator
            U_Ib = U[xdomain.o_Ipa][..., 0, :]
            U_Ia_list.append(np.einsum(
                '...a,ab->...b', U_Ib, DELTA_x_ab
            ))
            U_Lipa = U[xdomain.o_Lipa]
            U_Lipa_list.append(np.einsum(
                '...a,ab->...b', U_Lipa, DELTA_x_ab
            ))
            if len(U_Lipa) == 0:
                raise ValueError('no fields for variable %s' % self.var)
        # Save the displacements
        U_Ia = np.vstack(U_Ia_list)
        U_Lipa = np.vstack(U_Lipa_list)
 #       state_vec = np.hstack(state_fields)
        fname = 'U_step_%008.4f' % t
        target_file = os.path.join(
            self.dir, fname.replace('.', '_')
        ) + '.' + self.extension
        # , state_vec=state_vec)
        np.savez(target_file, U_Ia=U_Ia, U_Lipa=U_Lipa)
        self.add_file(target_file)


class Viz3DLattice(Viz3D):

    warp_factor = tr.Float(1.0, auto_set=False, enter_set=True)

    def setup(self):
        m = mlab

        npzfile = np.load(self.vis3d.lattice_file)
        X_Ia = npzfile['X_Ia']
        I_Li = npzfile['I_Li']
        Xm_Lia = npzfile['Xm_Lia']

        self.d = tvtk.PolyData(points=X_Ia, lines=I_Li)
        self.warp_vector = m.pipeline.warp_vector(self.d)
        self.warp_vector.filter.scale_factor = self.warp_factor

        tube = m.pipeline.tube(self.warp_vector,
                               tube_radius=self.tube_radius)
        self.lines = m.pipeline.surface(tube, color=(0.1, 0.1, 0.1))

        X_aiL = np.einsum('Lia->aiL', X_Ia[I_Li])
        Xm_aiL = np.einsum('Lia->aiL', Xm_Lia)
        XIm_aiL = np.concatenate(
            [
                np.einsum('iaL->aiL',
                          np.array([X_aiL[:, 0, ...], Xm_aiL[:, 0, ...]])),
                np.einsum('iaL->aiL',
                          np.array([X_aiL[:, 1, ...], Xm_aiL[:, 0, ...]]))
            ], axis=-1
        )
        XIm_aLi = np.einsum('aiL->aLi', XIm_aiL)
        XIm_aJ = XIm_aLi.reshape(3, -1)
        XIm_Ja = np.einsum('aJ->Ja', XIm_aJ)
        J_Li = np.arange(XIm_aJ.shape[1]).reshape(-1, 2)
        self.d_J = tvtk.PolyData(points=XIm_Ja, lines=J_Li)
        self.warp_vector_J = m.pipeline.warp_vector(self.d_J)
        self.warp_vector_J.filter.scale_factor = self.warp_factor

        tube_J = m.pipeline.tube(self.warp_vector_J,
                                 tube_radius=self.tube_radius)
        self.lines_J = m.pipeline.surface(tube_J, color=(0.1, 0.1, 0.1))
        glyph = m.pipeline.glyph(self.warp_vector_J)

        glyph.glyph.glyph_source.glyph_source = glyph.glyph.glyph_source.glyph_dict[
            'cube_source']
        glyph.glyph.glyph.range = np.array([0., 0.1])
        glyph.glyph.scale_mode = 'scale_by_vector'
        glyph.glyph.color_mode = 'color_by_vector'

    tube_radius = tr.Float(0.01)

    visible = tr.Property(tr.Bool)

    def _set_visible(self, visible):
        return
        self.d.visible = visible

    def _get_visible(self):
        return True  # self.d.visible

    def plot(self, vot):
        fe_domain = self.vis3d.tstep.fe_domain
        xdomain = fe_domain[0].xmodel
        dXm_Lia = xdomain.dXm_Lia

        idx = self.vis3d.tstep.hist.get_time_idx(vot)
        fname = self.vis3d.file_list[idx]
        npzfile = np.load(fname)
        U_Ia = npzfile['U_Ia']
        self.d.point_data.vectors = U_Ia

        U_Lipa = npzfile['U_Lipa']
        Um_trans_Lia, Phi_Lia = np.einsum('Lipa->pLia', U_Lipa)
        Um_rot_Lia = np.einsum('abc,...b,...c->...a',
                               EPS, Phi_Lia, dXm_Lia)
        Um_Lia = Um_trans_Lia + Um_rot_Lia
        Um_trans_aiL = np.einsum('Lia->aiL', Um_trans_Lia)
        Um_aiL = np.einsum('Lia->aiL', Um_Lia)
        UIm_aiL = np.concatenate(
            [
                np.einsum('iaL->aiL',
                          np.array([Um_trans_aiL[:, 0, ...], Um_aiL[:, 0, ...]])),
                np.einsum('iaL->aiL',
                          np.array([Um_trans_aiL[:, 1, ...], Um_aiL[:, 1, ...]]))
            ], axis=-1
        )
        UIm_aLi = np.einsum('aiL->aLi', UIm_aiL)
        UIm_aJ = UIm_aLi.reshape(3, -1)
        UIm_Ja = np.einsum('aJ->Ja', UIm_aJ)
        self.d_J.point_data.vectors = UIm_Ja




# Traits imports
from traits.api import \
    Enum, Float, \
    Instance, Trait, HasTraits, on_trait_change, \
    Dict, Property, cached_property, WeakRef, String, \
    Constant, List

from ibvpy.tmodel.mats_eval import IMATSEval, MATSEval
from ibvpy.mathkit.mfn.mfn_ndgrid.mfn_ndgrid import MFnNDGrid, GridPoint
#-------------------------------------------------------------------------
#                                     VariedParam
#-------------------------------------------------------------------------
class VariedParam(HasTraits):
    """
    Association between the spatial function and material parameter.
    """
    mats_eval = WeakRef(IMATSEval)

    varname = String

    reference_value = Property(Float, depends_on='mats_eval')

    @cached_property
    def _get_reference_value(self):
        return getattr(self.mats_eval, self.varname)

    switch = Enum('constant', 'varied')

    spatial_fn = Instance(MFnNDGrid)

    variable = Property(Float)

    def _get_variable(self):
        return getattr(self.mats_eval, self.varname)

    def _set_variable(self, value):
        setattr(self.mats_eval, self.varname, value)

    def adjust_for_sctx(self, sctx):
        if self.switch == 'varied':
            X_pnt = sctx.fets_eval.get_X_pnt(sctx)
            # X_coord = zeros(3)
            # X_coord[0:2] = X_pnt
            coeff = self.spatial_fn(X_pnt)
            self.variable = self.reference_value * coeff

#-------------------------------------------------------------------------
# Tabular Adapter Definition
#-------------------------------------------------------------------------

#---------------------------------------------------------------------------
# Material Proxy to include variable parameters
#---------------------------------------------------------------------------


class MATSProxy(MATSEval):
    '''
    Material model with spatially varying material parameters.

    @author: rch
    The material model works as a proxy delegating the standard
    functionality to the associated material model @param mats_eval.

    In the first step, the proxy identifies the material parameters
    of the mats_eval. The identification is performed by scanning the
    @param mats_eval for float traits using the method identify parameters.
    The scanning happens on demand when the @param
    varpars list is accessed.

    For each variable an instance of the VarPar object gets constructed
    to handle the spatial variation. By default, the VarPar has the
    switch set to constant so that the proxy has no effect.

    Spatial variation of a parameter within the VarPar instance
    can be activated both in the script during the model construction
    and interactively thorough the user interface. Both these ways are
    now briefly described to illuminate the internal workings
    of the proxy.

    Scripting interface
    ===================
    There is a quite number of dependencies involved in associating a
    spatial profile to a material parameter. Since this implementation
    is meant generic to be applicable to any material model,
    these dependcies must be captured in a transparent way. Let us
    identify the dependencies using a particular example.

    *** initial state ***

    proxy has an associated default material model.

    (@note: actually, it would be reasonable to reflect the usage
    context of the proxy, i.e. reuse the mats_eval used within of the
    element using the mats_proxy. Probably the spatial variation should intervene
    at a global level and hook up the spatial context anywhere upon the
    change of the spatial coordinate. In other words,
    the adjustment of a parameter in a particular time stepper would be done
    whenever the spatial coordinate gets changed.

    However, handling of these association
    is not the current priority. It can be solved by subclassing
    the proxy for the five supported dimensions, i.e. 1d, 1d5, 2d, 2d5 and 3d)

    upon access to varpar: identify the parameters of
    @param mats_eval and construct the VarPar instances. They are all
    set to constant.

    *** assign a spatial profile ***

    The variables must be accessible via their keywords. This means they are
    managed within a dictionary. For  instance you can issue

    mp = MATSProxy( mats_eval = MATS1DElastic() )
    mp.varpars['E'].spatial_fn.x_mins = [0,0,0]
    mp.varpars['E'].spatial_fn.x_maxs = [10,10,0]
    mp['E'].spatial_fn.shape = (500,5,0)
    mp['E'].spatial_fn.set_values_in_box( 10, [0,0,0], [10,10,8] )

    @todo The spatial variation can be handled at the level
    of the tstepper     Then, the identification of parameters
    can be performed for all integration levels. Further,
    the spatial binding can be taken into account.
    i.e., the spatial profile is specified for an existing
    domain with particular bounding box. The VaPars are registered
    in the spatial context and the adjustment
    of the material parameters is invoked dynamically,
    whenever the @param r_pnt gets changed. That means the correct
    parameters are available both for
    the iteration and for response tracing.

    The extensions needed -

    1) the recursive list of sub time steppers during the
        parameter identification in tstepper_eval

    2) spatial binding to the geometric model as a
        separate step (identification of the bounding box)

    3) spatial context is launched before starting the computation.
        at this point, the varpar_dict must be registered within
        the spatial context.

    3) spatial context must be hooked with the call to
       the adjust state variables.

    4) any change in the tstepper structure must
       cause reconstruction of the varpar list. Only those
       varpars should be modified that are really affected.
       (the need for change notification mechanism within the
       tstepper hierarchy)
    '''
#    mats_eval_type = Enum('MATS1DElastic',
#                          'MATS1DDamage',
#                          'MA2DCompositeMicroplaneDamage',
#                          'MATS2DScalarDamage')
#    mats_eval = Property( Instance( IMATSEval ), depends_on = 'mats_eval_type' )
#    @cached_property
#    def _get_mats_eval(self):
#        return eval( self.mats_eval_type + '()' )

#    def _set_mats_eval(self, value):
#        return value
    mats_eval = Instance(IMATSEval)
    #-------------------------------------------------------------------------
    # Management of spatially varying parameters depending on the value of mats_eval
    #-------------------------------------------------------------------------
    varpars = Dict

    def _varpars_default(self):
        return self._get_varpars()

    @on_trait_change('mats_eval')
    def _update_varpars(self):
        self.varpars = self._get_varpars()

    def _get_varpars(self):
        '''
        reset the varpar list according to the current mats_eval object.
        '''
        params = self.mats_eval.identify_parameters()
        varset = {}
        for key, par in list(params.items()):
            par_val = getattr(self.mats_eval, key)
            varset[key] = VariedParam(mats_eval=self.mats_eval,
                                      varname=key)
            #                                spatial_fn = 1 ) )
        return varset

    varpar_list = Property(List(VariedParam), depends_on='varpars')

    @cached_property
    def _get_varpar_list(self):
        return list(self.varpars.values())

    # variable selectable in the table of varied params (just for viewing)
    current_varpar = Instance(VariedParam)

    def _current_varpar_default(self):
        return self.varpar_list[0]

    @on_trait_change('mats_eval')
    def set_current_varpar(self):
        self.current_varpar = self.varpar_list[0]

    #-------------------------------------------------------------------------
    # Setup for computation within a supplied spatial context
    #-------------------------------------------------------------------------

    def get_state_array_size(self):
        return self.mats_eval.get_state_array_size()

    def setup(self, sctx):
        '''
        Intialize state variables.
        '''
        self.mats_eval.setup(sctx)

    def new_cntl_var(self):
        return self.mats_eval.new_cntl_var()

    def new_resp_var(self):
        return self.mats_eval.new_resp_var()

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, sctx, eps_app_eng, d_eps, tn, tn1, eps_avg=None):
        '''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        #
        # Reset the parameter values for the material model according to the
        # current spatial context
        #

        #        @todo check whether or not to put this into the spatial context - the X_pnt
        #        is needed only for certain configurations. If implemented as a cached functor
        #        it could be constructed on demand for all gauss points as a cached property
        # and the material models could reuse it without repeated evalution of
        # this operation.

        for varpar in self.varpar_list:
            varpar.adjust_for_sctx(sctx)

        return self.mats_eval.get_corr_pred(sctx, eps_app_eng, d_eps, tn, tn1, eps_avg)

    #-------------------------------------------------------------------------
    # Subsidiary methods realizing configurable features
    #-------------------------------------------------------------------------

    #-------------------------------------------------------------------------
    # Response trace evaluators
    #-------------------------------------------------------------------------

#    def get_eps_app( self, sctx, eps_app_eng ):
#        return eps_app_eng
#
#    def get_sig_app( self, sctx, eps_app_eng ):
#        # @TODO
#        # the stress calculation is performed twice - it might be
#        # cached.
#        sig_eng, D_mtx = self.get_corr_pred( sctx, eps_app_eng, 0, 0 )
#        return sig_eng

    # Declare and fill-in the rte_dict - it is used by the clients to
    # assemble all the available time-steppers.
    #
    rte_dict = Trait(Dict)

    def _rte_dict_default(self):
        return self.mats_eval.rte_dict



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Pierre Haessig — March 2014
"""
Qt adaptation of Gael Varoquaux's tutorial to integrate Matplotlib
http://docs.enthought.com/traitsui/tutorials/traits_ui_scientific_app.html#extending-traitsui-adding-a-matplotlib-figure-to-our-application

based on Qt-based code shared by Didrik Pinte, May 2012
http://markmail.org/message/z3hnoqruk56g2bje

adapted and tested to work with PySide from Anaconda in March 2014
"""

from matplotlib.backends.backend_qt5 import \
    NavigationToolbar2QT
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
from pyface.qt import QtGui
from traits.api import Instance
from traitsui.qt4.basic_editor_factory import BasicEditorFactory
from traitsui.qt4.editor import Editor


class _MPLFigureEditor(Editor):

    scrollable = True

    toolbar = True

    def init(self, parent):
        self.control = self._create_canvas(parent)
        self.object.on_trait_change(self.update_editor, 'data_changed')
        self.set_tooltip()

    def update_editor(self):
        figure = self.value
        figure.canvas.mpl_connect('key_press_event', self.key_press_callback)
        figure.canvas.draw()

    def _create_canvas(self, parent):
        """ Create the MPL canvas. """
        # matplotlib commands to create a canvas
        frame = QtGui.QWidget()
        mpl_canvas = FigureCanvas(self.value)
        mpl_canvas.setParent(frame)

        vbox = QtGui.QVBoxLayout()

        if self.toolbar:
            mpl_toolbar = NavigationToolbar2QT(mpl_canvas, frame)
            vbox.addWidget(mpl_toolbar)

        vbox.addWidget(mpl_canvas)
        frame.setLayout(vbox)

        return frame

    def key_press_callback(self, event):
        'whenever a key is pressed'
        figure = self.value
        if not event.inaxes:
            return
        if event.key == 'k':
            if figure.axes[0].get_xscale() == 'log':
                figure.axes[0].set_xscale('linear')
                figure.canvas.draw()
            else:
                figure.axes[0].set_xscale('log')
                figure.canvas.draw()

        if event.key == 'l':
            if figure.axes[0].get_yscale() == 'log':
                figure.axes[0].set_yscale('linear')
                figure.canvas.draw()
            else:
                figure.axes[0].set_yscale('log')
                figure.canvas.draw()


class MPLFigureEditor(BasicEditorFactory):

    klass = _MPLFigureEditor




from traits.api import \
    Instance, Array, Int, on_trait_change, Property, cached_property, \
    List, Button, provides,  \
    Event
from ibvpy.fets.i_fets import IFETSEval
from ibvpy.mesh.cell_grid.cell_array import ICellArraySource
from ibvpy.mesh.cell_grid.cell_spec import CellSpec
import numpy as np

from .fe_grid import FEGrid, MElem
from .fe_refinement_level import FERefinementLevel
from .i_fe_uniform_domain import IFEUniformDomain


@provides(ICellArraySource, IFEUniformDomain)
class FERefinementGrid(FERefinementLevel):

    '''Subgrid derived from another grid domain.
    '''

    changed_structure = Event

    @on_trait_change('+changed_structure')
    def set_changed_structure(self):
        self.changed_structure = True

    #-----------------------------------------------------------------
    # Feature: domain time-stepper
    #-----------------------------------------------------------------

    _fets_eval = Instance(IFETSEval)
    # inherit the fets_eval from the parent. This does not necessarily
    # have to be the case - calls for delegate here - that can be overloaded.
    fets_eval = Property

    def _set_fets_eval(self, value):
        self._fets_eval = value

    def _get_fets_eval(self):
        if self._fets_eval == None:
            return self.parent.fets_eval
        else:
            return self._fets_eval

    dof_r = Property

    def _get_dof_r(self):
        return self.fets_eval.dof_r

    geo_r = Property

    def _get_geo_r(self):
        return self.fets_eval.geo_r

    geo_transform = Property

    def _get_geo_transform(self):
        # geo transform should be part of a FEPatchedGrid
        #
        return self.parent.fe_subgrids[0].geo_transform

    n_nodal_dofs = Property

    def _get_n_nodal_dofs(self):
        return self.fets_eval.n_nodal_dofs

    # @TODO - move this to the base class - FEDomainBase
    #-------------------------------------------------------------------------
    # Derived properties
    #-------------------------------------------------------------------------
    # dof point distribution within the cell converted into the CellSpec format
    # CellSpec can derive the shape of the single grid cell, i.e.
    # the number of points specified in the individual directions.
    #
    dof_grid_spec = Property(Instance(CellSpec), depends_on='dof_r')

    def _get_dof_grid_spec(self):
        return CellSpec(node_coords=self.dof_r)

    # geo point distribution within the cell ... the same as above ...
    #
    geo_grid_spec = Property(Instance(CellSpec), depends_on='geo_r')

    def _get_geo_grid_spec(self):
        return CellSpec(node_coords=self.geo_r)

    fine_cell_shape = Array(int, value=[2, 2, 2])

    def get_fine_ix(self, coarse_ix):
        return np.array(list(coarse_ix), dtype=int) * self.fine_cell_shape

    def get_bounding_box(self, coarse_ix):
        '''Get the corner coordinates of the parent cell
        '''
        # @todo: TEMPORARY this must be done for the patched grid shape
        # FEPatchedGrid
        #
        pgrid = self.parent.fe_subgrids[0].geo_grid.point_x_grid
        print('shape', pgrid.shape)
        coarse_ix = np.array(list(coarse_ix), dtype=int)
        coord_min = pgrid[(slice(0, pgrid.shape[0]),) + tuple(coarse_ix)]
        coord_max = pgrid[(slice(0, pgrid.shape[0]),) + tuple(coarse_ix + 1)]
        return coord_min, coord_max

    def get_fine_fe_domain(self, coarse_ix):
        # get it for the general ix_slice
        # access fe_subgrid using coarse slices? or geometrically?
        # probably using fine slices?
        coord_min, coord_max = self.get_bounding_box(coarse_ix)
        fe_grid = FEGrid(fets_eval=self.fets_eval,
                         # multiply with the coarse ix / slice
                         shape=self.fine_cell_shape,
                         geo_transform=self.geo_transform,
                         coord_min=coord_min,
                         coord_max=coord_max)
        fe_grid.level = self
        return fe_grid

    def get_lset_subdomain(self, lset_function):
        '''@TODO - implement the subdomain selection method
        '''
        pass

    def get_boundary(self, side=None):
        '''@todo: - implement the boundary extraction
        '''
        pass

    def get_interior(self):
        '''@todo: - implement the boundary extraction
        '''
        pass

    #-------------------------------------------------------------
    # FEDomain interface
    #-------------------------------------------------------------

    elem_dof_map = Property(Array)

    def _get_elem_dof_map(self):
        elem_dof_map = np.vstack([fe_subgrid.elem_dof_map
                                  for fe_subgrid in self.fe_subgrids])
        return elem_dof_map

    elem_dof_map_unmasked = Property(Array)

    def _get_elem_dof_map_unmasked(self):
        elem_dof_map = np.vstack([fe_subgrid.elem_dof_map_unmasked
                                  for fe_subgrid in self.fe_subgrids])
        return elem_dof_map

    n_elems = Property

    def _get_n_elems(self):
        return len(self.elem_dof_map_unmasked)

    # get the number of dofs in the subgrids
    #  - consider caching
    n_dofs = Property(Int)

    def _get_n_dofs(self):
        '''Total number of dofs'''
        last_fe_subgrid = self.last_subgrid
        if last_fe_subgrid:
            return last_fe_subgrid.dof_offset + last_fe_subgrid.n_dofs - self.dof_offset
        else:
            return 0

    elem_X_map = Property(Array)

    def _get_elem_X_map(self):
        '''Array with the point coordinates'''
        return np.vstack([fe_subgrid.elem_X_map
                          for fe_subgrid in self.fe_subgrids])

    elem_X_map_unmasked = Property(Array)

    def _get_elem_X_map_unmasked(self):
        '''Array with the point coordinates'''
        return np.vstack([fe_subgrid.elem_X_map_unmasked
                          for fe_subgrid in self.fe_subgrids])

    elem_x_map = Property(Array)

    def _get_elem_x_map(self):
        '''Array with the point coordinates'''
        return np.vstack([fe_subgrid.elem_x_map
                          for fe_subgrid in self.fe_subgrids])

    elem_x_map_unmasked = Property(Array)

    def _get_elem_x_map_unmasked(self):
        '''Array with the point coordinates'''
        return np.vstack([fe_subgrid.elem_x_map_unmasked
                          for fe_subgrid in self.fe_subgrids])

    dof_Eid = Property
    '''Mapping of Element, Node, Dimension -> DOF 
    '''

    def _get_dof_Eid(self):
        return np.vstack([fe_subgrid.dof_Eid
                          for fe_subgrid in self.fe_subgrids])

    dofs = Property
    ''' 
    '''

    def _get_dofs(self):
        return np.vstack([fe_subgrid.dofs
                          for fe_subgrid in self.fe_subgrids])

    I_Ei = Property(Array)
    '''For a given element and its node number return the global index
    of the node'''

    def _get_I_Ei(self):
        return np.vstack([fe_subgrid.I_Ei
                          for fe_subgrid in self.fe_subgrids])

    X_Id = Property(Array)

    def _get_X_Id(self):
        return np.vstack([fe_subgrid.X_Id
                          for fe_subgrid in self.fe_subgrids])

    def deactivate(self, idx):
        '''Deactivate the specified element.

        The idx is an expanded cell index on the fine grid.
        Note, that the refined grid does not really exist
        as array structure, only the subgrids are there.
        Their offsets are set so as to fit into the implicit
        grid with the fineness defined by the current
        refinement level.
        '''
        # identify the correct subgrid to propagate the request to
        #
        # raise an index error if the idx points into non-existing element
        # (non-refined one)
        #
        for fe_grid in self.fe_subgrids:
            fe_grid.deactivate(idx)

    def reactivate(self, idx):
        '''Deactivate the specified element.

        The idx is an expanded cell index on the fine grid.
        Note, that the refined grid does not really exist
        as array structure, only the subgrids are there.
        Their offsets are set so as to fit into the implicit
        grid with the fineness defined by the current
        refinement level.
        '''
        raise NotImplementedError

#    fe_subgrids_params = Property
#    def _get_fe_subgrids_params( self ):
#        return zip( self.elem_dof_enumeration[6],
#                    self.elem_dof_enumeration[7],
#                    self.elem_dof_enumeration[5] )

    _fe_subgrids = List
    fe_subgrids = Property

    def _get_fe_subgrids(self):
        return self._fe_subgrids + self.elem_dof_enumeration[5]

    last_subgrid = Property

    def _get_last_subgrid(self):
        '''Return the last subgrids in order to establish the links.
        '''
        if len(self.fe_subgrids) > 0:
            return self.fe_subgrids[-1]
        else:
            return None

    def subgrids(self):
        #no_parents = [ None for i in range( len( self._fe_subgrids ) ) ]
        return list(zip(list(self.refinement_dict.keys()), self.fe_subgrids))

    elem_dof_enumeration = Property(
        depends_on='changed_structure,+changed_geometry')

    @cached_property
    def _get_elem_dof_enumeration(self):
        '''Array with the dof enumeration
        '''
        #p_list, args_list, fe_domain_list = [], [], []
        for p, refinement_args in list(self.refinement_dict.items()):

            fe_domain = self.get_fine_fe_domain(p)
            #fe_domain_list.append( fe_domain )
            #p_list.append( p )
            #args_list.append( refinement_args )

        prev_grid = None
        for fe_grid in self._fe_subgrids:
            if prev_grid:
                fe_grid.prev_grid = prev_grid
                prev_grid.next_grid = fe_grid
            prev_grid = fe_grid

        return [], -1, \
            [], [], [], \
            [], [], []
        #fe_domain_list, p_list, args_list

    #--------------------------------------------------------------
    # Activation should be implicit (include pending)
    #--------------------------------------------------------------
    # activation map
    #
    # get boolean array with inactive elements indicated by False
    activation_map = Property(depends_on='changed_structure')

    @cached_property
    def _get_activation_map(self):
        '''@TODO - react to changes in parent'''
        return np.hstack([subgrid.activation_map for subgrid in self.fe_subgrids])

    # get indices of all active elements
    idx_active_elems = Property(depends_on='changed_structure')

    @cached_property
    def _get_idx_active_elems(self):
        return np.arange(self.n_grid_elems)[self.activation_map]

    n_grid_elems = Property

    def _get_n_grid_elems(self):
        '''Total number of elements in the subgrid'''
        n_elem_arr = np.array(
            [subgrid.n_grid_elems for subgrid in self.fe_subgrids], dtype='int')
        return sum(n_elem_arr)

    n_active_elems = Property(
        List, depends_on='changed_structure,+changed_geometry,+changed_formulation,+changed_context')

    @cached_property
    def _get_n_active_elems(self):
        n_elem_arr = np.array(
            [subgrid.n_active_elems for subgrid in self.fe_subgrids], dtype='int')
        return sum(n_elem_arr)

    elements = Property(
        List, depends_on='changed_structure,+changed_geometry,+changed_formulation,+changed_context')

    @cached_property
    def _get_elements(self):
        '''The active list of elements to be included in the spatial integration'''
        # only active elements are returned
        return [MElem(dofs=dofs, point_X_arr=point_X_arr, point_x_arr=point_x_arr)
                for dofs, point_X_arr, point_x_arr in zip(self.elem_dof_map,
                                                          self.elem_X_map, self.elem_x_map)]

    def apply_on_ip_grid(self, fn, ip_mask):
        '''
        Apply the function fn over the first dimension of the array.
        @param fn: function to apply for each ip from ip_mask and each element.
        @param ip_mask: specifies the local coordinates within the element.
        '''
        X_el = self.elem_X_map
        # test call to the function with single output - to get the shape of
        # the result.
        out_single = fn(ip_mask[0], X_el[0])
        out_grid_shape = (X_el.shape[0], ip_mask.shape[0],) + out_single.shape
        out_grid = np.zeros(out_grid_shape)

        for el in range(X_el.shape[0]):
            for ip in range(ip_mask.shape[0]):
                out_grid[el, ip, ...] = fn(ip_mask[ip], X_el[el])

        return out_grid

    def apply_on_ip_grid_unmasked(self, fn, ip_mask):
        '''
        Apply the function fn over the first dimension of the array.
        @param fn: function to apply for each ip from ip_mask and each element.
        @param ip_mask: specifies the local coordinates within the element.
        '''
        X_el = self.elem_X_map_unmasked
        # test call to the function with single output - to get the shape of
        # the result.
        out_single = fn(ip_mask[0], X_el[0])
        out_grid_shape = (X_el.shape[0], ip_mask.shape[0],) + out_single.shape
        out_grid = np.zeros(out_grid_shape)

        for el in range(X_el.shape[0]):
            for ip in range(ip_mask.shape[0]):
                out_grid[el, ip, ...] = fn(ip_mask[ip], X_el[el])

        return out_grid

    #-------------------------------------------------------------
    # Visual introspection
    #-------------------------------------------------------------
    refresh_button = Button('Draw')

    @on_trait_change('refresh_button')
    def redraw(self):
        '''Redraw the point grid.
        @TODO
        '''
        for fe_grid in self.fe_subgrids:
            fe_grid.redraw()





from traits.api import HasTraits


class _MPLFigureEditor(HasTraits):

    def init(self, parent):
        pass

    def update_editor(self):
        pass


class MPLFigureEditor(HasTraits):

    klass = _MPLFigureEditor



'''
Created on 29.03.2017

@author: abaktheer

Microplane Fatigue model 2D

(compressive plasticity (CP) + tensile damage (TD) 
+ cumulative damage sliding (CSD))

Using Jirasek homogenization approach [1999]
'''

from ibvpy.tmodel.mats3D.mats3D_eval import MATS3DEval
from numpy import array,\
    einsum, zeros_like, identity, sign,\
    sqrt
from traits.api import Constant,\
    Property, cached_property

from bmcs_utils.api import Float

import numpy as np
import traits.api as tr


class MATS3DMplCSDODF(MATS3DEval):
    '''Tangential constitutive law parameters
    '''
    gamma_T = Float(5000.,
                    label="Gamma",
                    desc=" Tangential Kinematic hardening modulus",
                    enter_set=True,
                    auto_set=False)

    K_T = Float(10.0,
                label="K",
                desc="Tangential Isotropic harening",
                enter_set=True,
                auto_set=False)

    S = Float(0.00001,
              label="S",
              desc="Damage strength",
              enter_set=True,
              auto_set=False)

    r = Float(1.2,
              label="r",
              desc="Damage cumulation parameter",
              enter_set=True,
              auto_set=False)

    c = Float(1.0,
              label="c",
              desc="Damage cumulation parameter",
              enter_set=True,
              auto_set=False)

    tau_pi_bar = Float(2.0,
                       label="Tau_bar",
                       desc="Reversibility limit",
                       enter_set=True,
                       auto_set=False)

    a = Float(0.0,
              label="a",
              desc="Lateral pressure coefficient",
              enter_set=True,
              auto_set=False)

    #-------------------------------------------
    # Normal_Tension constitutive law parameters
    #-------------------------------------------
    Ad = Float(10000.0,
               label="a",
               desc="brittleness coefficient",
               enter_set=True,
               auto_set=False)

    eps_0 = Float(56e-6,
                  label="a",
                  desc="threshold strain",
                  enter_set=True,
                  auto_set=False)

    eps_f = Float(250e-6,
                  label="a",
                  desc="threshold strain",
                  enter_set=True,
                  auto_set=False)

    #-----------------------------------------------
    # Normal_Compression constitutive law parameters
    #-----------------------------------------------
    K_N = Float(10000.,
                label="K_N",
                desc=" Normal isotropic harening",
                enter_set=True,
                auto_set=False)

    gamma_N = Float(15000.,
                    label="gamma_N",
                    desc="Normal kinematic hardening",
                    enter_set=True,
                    auto_set=False)

    sigma_0 = Float(20.,
                    label="sigma_0",
                    desc="Yielding stress",
                    enter_set=True,
                    auto_set=False)

    zeta_G = Float(1.0,
                   label="zeta_G",
                   desc="anisotropy parameter",
                   enter_set=True,
                   auto_set=False)

    state_var_shapes = tr.Property(tr.Dict(), depends_on='n_mp')
    '''Dictionary of state variable entries with their array shapes.
    '''
    @cached_property
    def _get_state_shapes(self):
        return {'w_N_Emn': (self.n_mp,),
                'z_N_Emn': (self.n_mp,),
                'alpha_N_Emn': (self.n_mp,),
                'r_N_Emn': (self.n_mp,),
                'eps_N_p_Emn': (self.n_mp,),
                'sigma_N_Emn': (self.n_mp,),
                'w_T_Emn': (self.n_mp,),
                'z_T_Emn': (self.n_mp,),
                'alpha_T_Emna': (self.n_mp, 3),
                'eps_T_pi_Emna': (self.n_mp, 3), }

    #--------------------------------------------------------------
    # microplane constitutive law (normal behavior CP + TD)
    #--------------------------------------------------------------
    def get_normal_law(self, eps_N_Emn, w_N_Emn, z_N_Emn,
                       alpha_N_Emn, r_N_Emn, eps_N_p_Emn):

        E_N = self.E / (1.0 - 2.0 * self.nu)

        pos = eps_N_Emn > 1e-6
        H = 1.0 * pos

        sigma_n_trial = (1.0 - H * w_N_Emn) * E_N * (eps_N_Emn - eps_N_p_Emn)
        Z = self.K_N * r_N_Emn
        X = self.gamma_N * alpha_N_Emn

        h = self.sigma_0 + Z
        pos_iso = h > 1e-6
        f_trial = abs(sigma_n_trial - X) - h * pos_iso

        thres_1 = f_trial > 1e-6

        delta_lamda = f_trial / \
            (E_N + abs(self.K_N) + self.gamma_N) * thres_1
        eps_N_p_Emn = eps_N_p_Emn + delta_lamda * sign(sigma_n_trial - X)
        r_N_Emn = r_N_Emn + delta_lamda
        alpha_N_Emn = alpha_N_Emn + delta_lamda * sign(sigma_n_trial - X)

#         def Z_N(z_N_Emn): return 1.0 / self.Ad * (-z_N_Emn) / (1.0 + z_N_Emn)
#         Y_N = 0.5 * H * E_N * eps_N_Emn ** 2.0
#         Y_0 = 0.5 * E_N * self.eps_0 ** 2.0
#         f = Y_N - (Y_0 + Z_N(z_N_Emn))
#
#         thres_2 = f > 1e-6
#
#         def f_w(Y): return 1.0 - 1.0 / (1.0 + self.Ad * (Y - Y_0))
#         w_N_Emn = f_w(Y_N) * thres_2
#         z_N_Emn = - w_N_Emn * thres_2

        f_trial_Emn = eps_N_Emn - self.eps_0
        f_idx = np.where(f_trial_Emn > 0)
        z_N_Emn[f_idx] = eps_N_Emn[f_idx]
        w_N_Emn[f_idx] = (1. -
                          (self.eps_0 / z_N_Emn[f_idx]) * np.exp(- (z_N_Emn[f_idx] - self.eps_0) / (self.eps_f - self.eps_0)))

        sigma_N_Emn = (1.0 - H * w_N_Emn) * E_N * (eps_N_Emn - eps_N_p_Emn)

        return w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn

    #-------------------------------------------------------------------------
    # microplane constitutive law (Tangential CSD)-(Pressure sensitive cumulative damage)
    #-------------------------------------------------------------------------
    def get_tangential_law(self, eps_T_Emna, w_T_Emn, z_T_Emn,
                           alpha_T_Emna, eps_T_pi_Emna, sigma_kk):

        E_T = self.E / (1.0 + self.nu)

        sig_pi_trial = E_T * (eps_T_Emna - eps_T_pi_Emna)
        Z = self.K_T * z_T_Emn
        X = self.gamma_T * alpha_T_Emna
        norm_1 = sqrt(
            einsum('Emna,Emna -> Emn', (sig_pi_trial - X), (sig_pi_trial - X)))

        f = norm_1 - self.tau_pi_bar - \
            Z + self.a * sigma_kk

        plas_1 = f > 1e-6
        elas_1 = f < 1e-6

        delta_lamda = f / \
            (E_T / (1.0 - w_T_Emn) + self.gamma_T + self.K_T) * plas_1

#         print 'f', f
#         print 'w_T_Emn', w_T_Emn
#         print 'delta_lamda', delta_lamda
#         print '*****'

        norm_2 = 1.0 * elas_1 + sqrt(
            einsum('Emna,Emna -> Emn', (sig_pi_trial - X), (sig_pi_trial - X))) * plas_1

        eps_T_pi_Emna[..., 0] = eps_T_pi_Emna[..., 0] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 0] - X[..., 0]) /
             (1.0 - w_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 1] = eps_T_pi_Emna[..., 1] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 1] - X[..., 1]) /
             (1.0 - w_T_Emn)) / norm_2
        eps_T_pi_Emna[..., 2] = eps_T_pi_Emna[..., 2] + plas_1 * delta_lamda * \
            ((sig_pi_trial[..., 2] - X[..., 2]) /
             (1.0 - w_T_Emn)) / norm_2

        Y = 0.5 * E_T * \
            einsum('Emna,Emna -> Emn', (eps_T_Emna - eps_T_pi_Emna),
                   (eps_T_Emna - eps_T_pi_Emna))

        w_T_Emn += ((1.0 - w_T_Emn) ** self.c) * \
            (delta_lamda * (Y / self.S) ** self.r)  # * \
        #(self.tau_pi_bar / (self.tau_pi_bar - self.a * sigma_kk / 3.0))

        alpha_T_Emna[..., 0] = alpha_T_Emna[..., 0] + plas_1 * delta_lamda *\
            (sig_pi_trial[..., 0] - X[..., 0]) / norm_2
        alpha_T_Emna[..., 1] = alpha_T_Emna[..., 1] + plas_1 * delta_lamda *\
            (sig_pi_trial[..., 1] - X[..., 1]) / norm_2
        alpha_T_Emna[..., 2] = alpha_T_Emna[..., 2] + plas_1 * delta_lamda *\
            (sig_pi_trial[..., 2] - X[..., 2]) / norm_2
        z_T_Emn = z_T_Emn + delta_lamda

        return w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna

    #-------------------------------------------------------------------------
    # MICROPLANE-Kinematic constraints
    #-------------------------------------------------------------------------
    def _get_e_Emna(self, eps_Emab):
        # Projection of apparent strain onto the individual microplanes
        e_ni = einsum('nb,Emba->Emna', self._MPN, eps_Emab)
        return e_ni

    def _get_e_N_Emn(self, e_Emna):
        # get the normal strain array for each microplane
        e_N_Emn = einsum('Emna, na->Emn', e_Emna, self._MPN)
        return e_N_Emn

    def _get_e_T_Emna(self, e_Emna):
        # get the tangential strain vector array for each microplane
        e_N_Emn = self._get_e_N_Emn(e_Emna)
        e_N_Emna = einsum('Emn,na->Emna', e_N_Emn, self._MPN)
        return e_Emna - e_N_Emna

    #-------------------------------------------------
    # Alternative methods for the kinematic constraint
    #-------------------------------------------------
    # get the dyadic product of the microplane normals
    _MPNN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPNN(self):
        MPNN_nij = einsum('ni,nj->nij', self._MPN, self._MPN)
        return MPNN_nij

    # get the third order tangential tensor (operator) for each microplane
    _MPTT = Property(depends_on='n_mp')

    @cached_property
    def _get__MPTT(self):
        delta = identity(3)
        MPTT_nijr = 0.5 * (einsum('ni,jr -> nijr', self._MPN, delta) +
                           einsum('nj,ir -> njir', self._MPN, delta) - 2 *
                           einsum('ni,nj,nr -> nijr', self._MPN, self._MPN, self._MPN))
        return MPTT_nijr

    def _get_e_N_Emn_2(self, eps_Emab):
        # Projection of apparent strain onto the individual microplanes
        return einsum('nij,Emij->Emn', self._MPNN, eps_Emab)

    def _get_e_T_Emna_2(self, eps_Emab):
        # get the normal strain array for each microplane
        MPTT_ijr = self._get__MPTT()
        return einsum('nija,Emij->Emna', MPTT_ijr, eps_Emab)

    def _get_e_Emna_2(self, eps_Emab):
        # get the tangential strain vector array for each microplane
        return self._get_e_N_Emn_2(eps_Emab) * self._MPN +\
            self._get_e_T_Emna_2(eps_Emab)

    #--------------------------------------------------------
    # return the state variables (Damage , inelastic strains)
    #--------------------------------------------------------
    def _get_state_variables(self, sctx, eps_app_eng, sigma_kk):

        e_N_arr = self._get_e_N_arr_2(eps_app_eng)
        e_T_vct_arr = self._get_e_T_vct_arr_2(eps_app_eng)

        sctx_arr = zeros_like(sctx)

        sctx_N = self.get_normal_law(e_N_arr, sctx)
        sctx_arr[:, 0:5] = sctx_N

        sctx_tangential = self.get_tangential_law(e_T_vct_arr, sctx, sigma_kk)
        sctx_arr[:, 5:13] = sctx_tangential

        return sctx_arr

    #-----------------------------------------------------------------
    # Returns a list of the plastic normal strain  for all microplanes.
    #-----------------------------------------------------------------
    def _get_eps_N_p_Emn(self, eps_Emab, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn):
        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_N_p_Emn = self.get_normal_law(
            eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[4]
        return eps_N_p_Emn

    #----------------------------------------------------------------
    # Returns a list of the sliding strain vector for all microplanes.
    #----------------------------------------------------------------
    def _get_eps_T_pi_Emna(self, eps_Emab, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_kk_Em):

        eps_T_Emna = self._get_e_T_Emna_2(eps_Emab)
        eps_N_T_pi_Emna = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_kk_Em)[3]

        return eps_N_T_pi_Emna

    #-------------------------------------------------------------
    # Returns a list of the integrity factors for all microplanes.
    #-------------------------------------------------------------
    def _get_phi_Emn(self, eps_Emab, w_N_Emn, z_N_Emn,
                     alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                     w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna,  sigma_N_Emn):

        eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
        eps_T_Emna = self._get_e_T_Emna_2(eps_Emab)

        w_N_Emn = self.get_normal_law(
            eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[0]
        sigma_N_Emn = self.get_normal_law(
            eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[5]
        w_T_Emn = self.get_tangential_law(
            eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)[0]

        w_Emn = zeros_like(w_N_Emn)

#         w_Emn = np.zeros_like(w_N_Emn)
#         idx_1 = np.where(sigma_N_Emn >= 0)
#         w_Emn[idx_1] = w_N_Emn[idx_1]
#
#         idx_2 = np.where(sigma_N_Emn < 0)
#         w_Emn[idx_2] = w_T_Emn[idx_2]
#
        eig = np.linalg.eig(eps_Emab)[0]

        ter_1 = np.sum(eig)
        idx_1 = np.where(ter_1 > 0.0)
        idx_2 = np.where(ter_1 <= 0.0)

        w_Emn[idx_1] = w_N_Emn[idx_1]
        w_Emn[idx_2] = w_T_Emn[idx_2]

        phi_Emn = 1.0 - w_Emn

        return phi_Emn

    #----------------------------------------------
    # Returns the 2nd order damage tensor 'phi_mtx'
    #----------------------------------------------
    def _get_phi_Emab(self, phi_Emn):

        # integration terms for each microplanes
        phi_Emab = einsum('Emn,n,nab->Emab', phi_Emn, self._MPW, self._MPNN)

        return phi_Emab
#
#     #----------------------------------------------------------------------
#     # Returns the 4th order damage tensor 'beta4' using sum-type symmetrization
#     #(cf. [Jir99], Eq.(21))
#     #----------------------------------------------------------------------
#     def _get_beta_Emabcd(self, phi_Emab):
#
#         delta = identity(3)
#
#         # use numpy functionality (einsum) to evaluate [Jir99], Eq.(21)
#         beta_Emabcd = 0.25 * (einsum('Emik,jl->Emijkl', phi_Emab, delta) +
#                               einsum('Emil,jk->Emijkl', phi_Emab, delta) +
#                               einsum('Emjk,il->Emijkl', phi_Emab, delta) +
#                               einsum('Emjl,ik->Emijkl', phi_Emab, delta))
#
#         return beta_Emabcd
#
# #     #----------------------------------------------------------------------
# #     # Returns the 4th order damage tensor 'beta4' using product-type symmetrization
# #     #(cf. [Baz97], Eq.(87))
# #     #----------------------------------------------------------------------
# #     def _get_beta_tns_product_type(self, sctx, eps_app_eng, sigma_kk):
# #
# #         delta = identity(2)
# #
# #         phi_mtx = self._get_phi_mtx(sctx, eps_app_eng, sigma_kk)
# #
# #         n_dim = 2
# #         phi_eig_value, phi_eig_mtx = eigh(phi_mtx)
# #         phi_eig_value_real = array([pe.real for pe in phi_eig_value])
# #         phi_pdc_mtx = zeros((n_dim, n_dim), dtype=float)
# #         for i in range(n_dim):
# #             phi_pdc_mtx[i, i] = phi_eig_value_real[i]
# #         # w_mtx = tensorial square root of the second order damage tensor:
# #         w_pdc_mtx = sqrt(phi_pdc_mtx)
# #
# #         # transform the matrix w back to x-y-coordinates:
# #         w_mtx = einsum('ik,kl,lj -> ij', phi_eig_mtx, w_pdc_mtx, phi_eig_mtx)
# #         #w_mtx = dot(dot(phi_eig_mtx, w_pdc_mtx), transpose(phi_eig_mtx))
# #
# #         beta_ijkl = 0.5 * \
# #             (einsum('ik,jl -> ijkl', w_mtx, w_mtx) +
# #              einsum('il,jk -> ijkl', w_mtx, w_mtx))
# #
# #         return beta_ijkl
#
#     #-----------------------------------------------------------
#     # Integration of the (inelastic) strains for each microplane
#     #-----------------------------------------------------------
#     def _get_eps_p_Emab(self, eps_Emab, w_N_Emn, z_N_Emn,
#                         alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
#                         w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna,  sigma_N_Emn):
#
#         eps_N_Emn = self._get_e_N_Emn_2(eps_Emab)
#         eps_T_Emna = self._get_e_T_Emna_2(eps_Emab)
#
# #         # plastic normal strains
# #         eps_N_p_Emn = self.get_normal_law(
# # eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)[5]
#
#         eps_N_p_Emn = self._get_eps_N_p_Emn(
#             eps_Emab, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)
#
# #         # sliding tangential strains
# #         eps_T_pi_Emna = self.get_tangential_law(
# # eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna,
# # sigma_N_Emn)[4]
#
#         eps_T_pi_Emna = self._get_eps_T_pi_Emna(
#             eps_Emab, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)
#
#         delta = identity(3)
#
#         # 2-nd order plastic (inelastic) tensor
#         eps_p_Emab = einsum('n,Emn,na,nb -> Emab', self._MPW, eps_N_p_Emn, self._MPN, self._MPN) + \
#             0.5 * (einsum('n,Emnf,na,fb-> Emab', self._MPW, eps_T_pi_Emna, self._MPN, delta) +
#                    einsum('n,Emnf,nb,fa-> Emab', self._MPW, eps_T_pi_Emna, self._MPN, delta))
#
#         return eps_p_Emab
#
#     #-------------------------------------------------------------------------
#     # Evaluation - get the corrector and predictor
#     #-------------------------------------------------------------------------
#
#     def get_corr_pred(self, eps_Emab_n1, deps_Emab, tn, tn1, update_state, w_N_Emn, z_N_Emn,
#                       alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn,
#                       w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna):
#
#         # Corrector predictor computation.
#         if update_state:
#
#             eps_Emab_n = eps_Emab_n1 - deps_Emab
#
#             eps_N_Emn = self._get_e_N_Emn_2(eps_Emab_n)
#             eps_T_Emna = self._get_e_T_Emna_2(eps_Emab_n)
#
#             w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn = self.get_normal_law(
#                 eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)
#             w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna = self.get_tangential_law(
#                 eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)
#
#         #------------------------------------------------------------------
#         # Damage tensor (4th order) using product- or sum-type symmetrization:
#         #------------------------------------------------------------------
#         phi_Emn = self._get_phi_Emn(eps_Emab_n1, w_N_Emn, z_N_Emn,
#                                     alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
#                                     w_T_Emn, z_T_Emn, alpha_T_Emna,
#                                     eps_T_pi_Emna, sigma_N_Emn)
#
#         phi_Emab = self._get_phi_Emab(phi_Emn)
#
#         beta_Emabcd = self._get_beta_Emabcd(phi_Emab)
#
#         #------------------------------------------------------------------
#         # Damaged stiffness tensor calculated based on the damage tensor beta4:
#         #------------------------------------------------------------------
#         D_Emabcd = einsum(
#             'Emijab, abef, Emcdef -> Emijcd', beta_Emabcd, self.D_abef, beta_Emabcd)
#         #----------------------------------------------------------------------
#         # Return stresses (corrector) and damaged secant stiffness matrix (predictor)
#         #----------------------------------------------------------------------
#         # plastic strain tensor
#         eps_p_Emab = self._get_eps_p_Emab(eps_Emab_n1, w_N_Emn, z_N_Emn,
#                                           alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
#                                           w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)
#
#         # elastic strain tensor
#         eps_e_Emab = eps_Emab_n1 - eps_p_Emab
#
#         # calculation of the stress tensor
#         sig_Emab = einsum('Emabcd,Emcd -> Emab', D_Emabcd, eps_e_Emab)
#
#         return D_Emabcd, sig_Emab

    #----------------------------------------------------------------
    #  the fourth order volumetric-identity tensor
    #----------------------------------------------------------------
    def _get_I_vol_abcd(self):

        delta = identity(3)
        I_vol_abcd = (1.0 / 3.0) * einsum('ab,cd -> abcd', delta, delta)
        return I_vol_abcd

    #----------------------------------------------------------------
    # Returns the fourth order deviatoric-identity tensor
    #----------------------------------------------------------------
    def _get_I_dev_abcd(self):

        delta = identity(3)
        I_dev_abcd = 0.5 * (einsum('ac,bd -> abcd', delta, delta) +
                            einsum('ad,bc -> abcd', delta, delta)) \
            - (1.0 / 3.0) * einsum('ab,cd -> abcd', delta, delta)

        return I_dev_abcd

    #----------------------------------------------------------------
    # Returns the fourth order tensor P_vol [Wu.2009]
    #----------------------------------------------------------------
    def _get_P_vol_ab(self):

        delta = identity(3)
        P_vol_ab = (1.0 / 3.0) * delta
        return P_vol_ab

    #----------------------------------------------------------------
    # Returns the fourth order tensor P_dev [Wu.2009]
    #----------------------------------------------------------------
    def _get_P_dev_nabc(self):

        delta = identity(3)
        P_dev_nabc = 0.5 * einsum('nd,da,bc -> nabc', self._MPN, delta, delta)
        return P_dev_nabc

    #----------------------------------------------------------------
    # Returns the outer product of P_vol [Wu.2009]
    #----------------------------------------------------------------
    def _get_PP_vol_abcd(self):

        delta = identity(3)
        PP_vol_abcd = (1.0 / 9.0) * einsum('ab,cd -> abcd', delta, delta)
        return PP_vol_abcd

    #----------------------------------------------------------------
    # Returns the inner product of P_dev
    #----------------------------------------------------------------
    def _get_PP_dev_nabcd(self):

        delta = identity(3)
        PP_dev_nabcd = 0.5 * (0.5 * (einsum('na,nc,bd -> nabcd', self._MPN, self._MPN, delta) +
                                     einsum('na,nd,bc -> nabcd', self._MPN, self._MPN, delta)) +
                              0.5 * (einsum('ac,nb,nd -> nabcd',  delta, self._MPN, self._MPN) +
                                     einsum('ad,nb,nc -> nabcd',  delta, self._MPN, self._MPN))) -\
            (1.0 / 3.0) * (einsum('na,nb,cd -> nabcd', self._MPN, self._MPN, delta) +
                           einsum('ab,nc,nd -> nabcd', delta, self._MPN, self._MPN)) +\
            (1.0 / 9.0) * einsum('ab,cd -> abcd', delta, delta)

        return PP_dev_nabcd

    def _get_I_vol_4(self):
        # The fourth order volumetric-identity tensor
        delta = identity(3)
        I_vol_ijkl = (1.0 / 3.0) * einsum('ij,kl -> ijkl', delta, delta)
        return I_vol_ijkl

    def _get_I_dev_4(self):
        # The fourth order deviatoric-identity tensor
        delta = identity(3)
        I_dev_ijkl = 0.5 * (einsum('ik,jl -> ijkl', delta, delta) +
                            einsum('il,jk -> ijkl', delta, delta)) \
            - (1 / 3.0) * einsum('ij,kl -> ijkl', delta, delta)

        return I_dev_ijkl

    def _get_P_vol(self):
        delta = identity(3)
        P_vol_ij = (1 / 3.0) * delta
        return P_vol_ij

    def _get_P_dev(self):
        delta = identity(3)
        P_dev_njkl = 0.5 * einsum('ni,ij,kl -> njkl', self._MPN, delta, delta)
        return P_dev_njkl

    def _get_PP_vol_4(self):
        # outer product of P_vol
        delta = identity(3)
        PP_vol_ijkl = (1 / 9.) * einsum('ij,kl -> ijkl', delta, delta)
        return PP_vol_ijkl

    def _get_PP_dev_4(self):
        # inner product of P_dev
        delta = identity(3)
        PP_dev_nijkl = 0.5 * (0.5 * (einsum('ni,nk,jl -> nijkl', self._MPN, self._MPN, delta) +
                                     einsum('ni,nl,jk -> nijkl', self._MPN, self._MPN, delta)) +
                              0.5 * (einsum('ik,nj,nl -> nijkl',  delta, self._MPN, self._MPN) +
                                     einsum('il,nj,nk -> nijkl',  delta, self._MPN, self._MPN))) -\
            (1 / 3.) * (einsum('ni,nj,kl -> nijkl', self._MPN, self._MPN, delta) +
                        einsum('ij,nk,nl -> nijkl', delta, self._MPN, self._MPN)) +\
            (1 / 9.) * einsum('ij,kl -> ijkl', delta, delta)
        return PP_dev_nijkl

    #--------------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor (cf. [Wu.2009], Eq.(29))
    #--------------------------------------------------------------------------
    def _get_S_1_Emabcd(self, eps_Emab, phi_Emn):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)

        #phi_Emn = 1.0 - self._get_omega(kappa)
        # print 'phi_Emn', phi_Emn

        PP_vol_abcd = self._get_PP_vol_abcd()
        PP_dev_nabcd = self._get_PP_dev_nabcd()
        I_dev_abcd = self._get_I_dev_abcd()

#         PP_vol_abcd = self._get_PP_vol_4()
#         PP_dev_nabcd = self._get_PP_dev_4()
#         I_dev_abcd = self._get_I_dev_4()

        S_1_Emabcd = K0 * einsum('Emn, n, abcd-> Emabcd', phi_Emn, self._MPW, PP_vol_abcd) + \
            G0 * 2.0 * self.zeta_G * einsum('Emn, n, nabcd-> Emabcd',
                                            phi_Emn, self._MPW, PP_dev_nabcd) - (1.0 / 3.0) * (
                2.0 * self.zeta_G - 1.0) * G0 * einsum('Emn, n, abcd-> Emabcd',
                                                       phi_Emn, self._MPW, I_dev_abcd)

        return S_1_Emabcd

#     #------------------------------------------
#     # scalar damage factor for each microplane
#     #------------------------------------------
#     def _get_d_Em(self, s_Emng, eps_Emab):
#
#         d_Emn = 1.0 - self.get_state_variables(s_Emng, eps_Emab)[0]
#
#         d_Em = (1.0 / 3.0) * einsum('Emn,n-> Em',  d_Emn, self._MPW)
#
#         return d_Em
#
#     #------------------------------------------
#     # The 4th order volumetric damage tensor
#     #------------------------------------------
#     def _get_M_vol_abcd(self, sctx, eps_app_eng, sigma_kk):
#
#         d = self._get_Em( s_Emng, eps_Emab)
#         delta = identity(2)
#
#         I_4th_abcd = 0.5 * (einsum('ac,bd -> ijkl', delta, delta) +
#                             einsum('il,jk -> ijkl', delta, delta))
#
#         # print 'M_vol', (1 - d) * I_4th_ijkl
#
#         return (1 - d) * I_4th_ijkl
#
#     #------------------------------------------
#     # The 4th order deviatoric damage tensor
#     #------------------------------------------
#     def _get_M_dev_tns(self, phi_mtx):
#
#         delta = identity(3)
#         I_4th_ijkl = 0.5 * (einsum('ik,jl -> ijkl', delta, delta) +
#                             einsum('il,jk -> ijkl', delta, delta))
#         tr_phi_mtx = np.trace(phi_mtx)
#
#         M_dev_ijkl = self.zeta_G * (0.5 * (einsum('ik,jl->ijkl', delta, phi_mtx) +
#                                            einsum('il,jk->ijkl', delta, phi_mtx)) +
#                                     0.5 * (einsum('ik,jl->ijkl', phi_mtx, delta) +
#                                            einsum('il,jk->ijkl', phi_mtx, delta))) \
#             - (2. * self.zeta_G - 1.) * (tr_phi_mtx / 3.) * I_4th_ijkl
#
#         return M_dev_ijkl
#
#     #--------------------------------------------------------------------------
#     # Returns the fourth order secant stiffness tensor (cf. [Wu.2009], Eq.(31))
#     #--------------------------------------------------------------------------
#     def _get_S_2_Emabcd(self, sctx, eps_app_eng, sigma_kk):
#
#         K0 = self.E / (1. - 2. * self.nu)
#         G0 = self.E / (1. + self.nu)
#
#         I_vol_ijkl = self._get_I_vol_4()
#         I_dev_ijkl = self._get_I_dev_4()
#         phi_mtx = self._get_phi_mtx(sctx, eps_app_eng, sigma_kk)
#         M_vol_ijkl = self._get_M_vol_tns(sctx, eps_app_eng, sigma_kk)
#         M_dev_ijkl = self._get_M_dev_tns(phi_mtx)
#
#         S_2_ijkl = K0 * einsum('ijmn,mnrs,rskl -> ijkl', I_vol_ijkl, M_vol_ijkl, I_vol_ijkl ) \
#             + G0 * einsum('ijmn,mnrs,rskl -> ijkl', I_dev_ijkl, M_dev_ijkl, I_dev_ijkl)\
#
#         return S_2_ijkl
#
#     #--------------------------------------------------------------------------
#     # Returns the fourth order secant stiffness tensor (cf. [Wu.2009], Eq.(34))
#     #--------------------------------------------------------------------------
#     def _get_S_3_Emabcd(self, sctx, eps_app_eng, sigma_kk):
#
#         K0 = self.E / (1. - 2. * self.nu)
#         G0 = self.E / (1. + self.nu)
#
#         I_vol_ijkl = self._get_I_vol_4()
#         I_dev_ijkl = self._get_I_dev_4()
#
#         # The fourth order elastic stiffness tensor
#         S_0_ijkl = K0 * I_vol_ijkl + G0 * I_dev_ijkl
#
#         d_n = self._get_state_variables(sctx, eps_app_eng, sigma_kk)[:, 5]
#
#         PP_vol_4 = self._get_PP_vol_4()
#         PP_dev_4 = self._get_PP_dev_4()
#
#         delta = identity(3)
#         I_4th_ijkl = einsum('ik,jl -> ijkl', delta, delta)
#
#         D_ijkl = einsum('n,n,ijkl->ijkl', d_n, self._MPW, PP_vol_4) + \
#             2 * self.zeta_G * einsum('n,n,nijkl->ijkl', d_n, self._MPW, PP_dev_4) - (
#                 1 / 3.) * (2 * self.zeta_G - 1) * einsum('n,n,ijkl->ijkl', d_n, self._MPW, I_dev_ijkl)
#
#         phi_ijkl = (I_4th_ijkl - D_ijkl)
#
#         S_ijkl = einsum('ijmn,mnkl', phi_ijkl, S_0_ijkl)
#
#         return S_ijkl
#
    #-------------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor with the (double orthotropic) assumption
    #-------------------------------------------------------------------------
    def _get_S_4_Emabcd(self, eps_Emab, phi_Emab):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)
        I_vol_abcd = self._get_I_vol_abcd()
        I_dev_abcd = self._get_I_dev_abcd()

        delta = identity(3)
        #phi_Emab = self._get_phi_Emab(kappa)

        D_Emab = delta - phi_Emab

        d_Em = (1.0 / 3.0) * np.einsum('Emaa -> Em', D_Emab)

        D_bar_Emab = self.zeta_G * \
            (D_Emab - np.einsum('Em, ab -> Emab', d_Em, delta))

        S_4_Emabcd = K0 * I_vol_abcd - K0 * einsum('Em,abcd -> Emabcd', d_Em, I_vol_abcd) +\
            G0 * I_dev_abcd - G0 * einsum('Em,abcd -> Emabcd', d_Em, I_dev_abcd) +\
            (2.0 / 3.0) * (G0 - K0) * (einsum('ij,Emkl -> Emijkl', delta, D_bar_Emab) +
                                       einsum('Emij,kl -> Emijkl', D_bar_Emab, delta)) + 0.5 * (- K0 + 2.0 * G0) *\
            (0.5 * (einsum('ik,Emjl -> Emijkl', delta, D_bar_Emab) + einsum('Emil,jk -> Emijkl', D_bar_Emab, delta)) +
             0.5 * (einsum('Emil,jk -> Emijkl', D_bar_Emab, delta) + einsum('ik,Emjl -> Emijkl', delta, D_bar_Emab)))

        return S_4_Emabcd

    #----------------------------------------------------------------------
    # Returns the fourth order secant stiffness tensor (restrctive orthotropic)
    #----------------------------------------------------------------------
    def _get_S_5_Emabcd(self, eps_Emab, phi_Emab):

        K0 = self.E / (1.0 - 2.0 * self.nu)
        G0 = self.E / (1.0 + self.nu)

        delta = identity(3)
        # commented out - @rch, @abdul please check
        # phi_Emab = self._get_phi_Emab(kappa)

        # damaged stiffness without simplification
        S_5_Emabcd = (1.0 / 3.0) * (K0 - G0) * 0.5 * ((einsum('ij,Emkl -> Emijkl', delta, phi_Emab) +
                                                       einsum('Emij,kl -> Emijkl', phi_Emab, delta))) + \
            G0 * 0.5 * ((0.5 * (einsum('ik,Emjl -> Emijkl', delta, phi_Emab) + einsum('Emil,jk -> Emijkl', phi_Emab, delta)) +
                         0.5 * (einsum('Emik,jl -> ijkl', phi_Emab, delta) + einsum('il,Emjk  -> Emijkl', delta, phi_Emab))))

        return S_5_Emabcd

    #-------------------------------------------------------------------------
    # Evaluation - get the corrector and predictor
    #-------------------------------------------------------------------------

    def get_corr_pred(self, eps_Emab_n1, deps_Emab, tn, tn1, update_state, w_N_Emn, z_N_Emn,
                      alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn,
                      w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna):

        if update_state:

            eps_Emab_n = eps_Emab_n1 - deps_Emab

            eps_N_Emn = self._get_e_N_Emn_2(eps_Emab_n)
            eps_T_Emna = self._get_e_T_Emna_2(eps_Emab_n)

            w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn, sigma_N_Emn = self.get_normal_law(
                eps_N_Emn, w_N_Emn, z_N_Emn, alpha_N_Emn, r_N_Emn, eps_N_p_Emn)
            w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna = self.get_tangential_law(
                eps_T_Emna, w_T_Emn, z_T_Emn, alpha_T_Emna, eps_T_pi_Emna, sigma_N_Emn)

        phi_Emn = self._get_phi_Emn(eps_Emab_n1, w_N_Emn, z_N_Emn,
                                    alpha_N_Emn, r_N_Emn, eps_N_p_Emn,
                                    w_T_Emn, z_T_Emn, alpha_T_Emna,
                                    eps_T_pi_Emna, sigma_N_Emn)

        phi_Emab = self._get_phi_Emab(phi_Emn)

        D_Emabcd = self._get_S_4_Emabcd(eps_Emab_n1, phi_Emab)

        sig_Emab = einsum('Emabcd,Emcd -> Emab', D_Emabcd, eps_Emab_n1)

        return D_Emabcd, sig_Emab

    #-----------------------------------------------
    # number of microplanes - currently fixed for 3D
    #-----------------------------------------------
    n_mp = Constant(28)

    #-----------------------------------------------
    # get the normal vectors of the microplanes
    #-----------------------------------------------
    _MPN = Property(depends_on='n_mp')

    @cached_property
    def _get__MPN(self):
        return array([[.577350259, .577350259, .577350259],
                      [.577350259, .577350259, -.577350259],
                      [.577350259, -.577350259, .577350259],
                      [.577350259, -.577350259, -.577350259],
                      [.935113132, .250562787, .250562787],
                      [.935113132, .250562787, -.250562787],
                      [.935113132, -.250562787, .250562787],
                      [.935113132, -.250562787, -.250562787],
                      [.250562787, .935113132, .250562787],
                      [.250562787, .935113132, -.250562787],
                      [.250562787, -.935113132, .250562787],
                      [.250562787, -.935113132, -.250562787],
                      [.250562787, .250562787, .935113132],
                      [.250562787, .250562787, -.935113132],
                      [.250562787, -.250562787, .935113132],
                      [.250562787, -.250562787, -.935113132],
                      [.186156720, .694746614, .694746614],
                      [.186156720, .694746614, -.694746614],
                      [.186156720, -.694746614, .694746614],
                      [.186156720, -.694746614, -.694746614],
                      [.694746614, .186156720, .694746614],
                      [.694746614, .186156720, -.694746614],
                      [.694746614, -.186156720, .694746614],
                      [.694746614, -.186156720, -.694746614],
                      [.694746614, .694746614, .186156720],
                      [.694746614, .694746614, -.186156720],
                      [.694746614, -.694746614, .186156720],
                      [.694746614, -.694746614, -.186156720]])

    #-------------------------------------
    # get the weights of the microplanes
    #-------------------------------------
    _MPW = Property(depends_on='n_mp')

    @cached_property
    def _get__MPW(self):
        return array([.0160714276, .0160714276, .0160714276, .0160714276, .0204744730,
                      .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                      .0204744730, .0204744730, .0204744730, .0204744730, .0204744730,
                      .0204744730, .0158350505, .0158350505, .0158350505, .0158350505,
                      .0158350505, .0158350505, .0158350505, .0158350505, .0158350505,
                      .0158350505, .0158350505, .0158350505]) * 6.0



'''
Created on 12.01.2016

@author: Yingxiong
'''
from traits.api import Int, HasStrictTraits, Instance, \
    Float, Event, \
    Array, List, Bool, Property, cached_property, Str, Dict

from ibvpy.sim.tline import TLine
from ibvpy.mathkit.linalg.sys_mtx_assembly import SysMtxAssembly
import numpy as np
from .mats3D.mats3D_explore import MATS3DExplore


class TLoop(HasStrictTraits):

    ts = Instance(MATS3DExplore)

    tline = Instance(TLine, ())
    d_t = Float(0.005)
    t_max = Float(1.0)
    k_max = Int(50)
    tolerance = Float(1e-4)
    step_tolerance = Float(1e-8)

    t_record = List
    U_n = Array(dtype=np.float64)
    K = Array(dtype=np.float64)
    state = Array(dtype=np.float64)
    F_record = List
    U_record = List
    state_record = List

    K = Instance(SysMtxAssembly)

    paused = Bool(False)
    restart = Bool(True)

    def setup(self):
        n_comps = 6
        n_dofs = n_comps
        self.U_n = np.zeros((n_dofs,))
        t_n = self.tline.val
        self.t_record = [t_n]
        self.U_record = [np.zeros_like(self.U_n)]
        self.F_record = [np.copy(self.U_n)]
        self.state_record = [np.copy(self.state)]
        # Set up the system matrix
        #
        self.K = SysMtxAssembly()
        self.ts.bcond_mngr.apply_essential(self.K)

    state_changed = Event
    state_arrays = Property(Dict(Str, Array),
                            depends_on='state_changed')
    '''Dictionary of state arrays.
    The entry names and shapes are defined by the material
    model.
    '''
    @cached_property
    def _get_state_arrays(self):
        sa_shapes = self.ts.state_array_shapes
        print('state array generated', sa_shapes)
        return {
            name: np.zeros(mats_sa_shape, dtype=np.float64)[np.newaxis, ...]
            for name, mats_sa_shape
            in list(sa_shapes.items())
        }

    def init(self):
        if self.paused:
            self.paused = False
        if self.restart:
            self.tline.val = 0
            self.state_changed = True
            self.setup()
            self.restart = False

    def eval(self):
        # Reset the system matrix (constraints are preserved)
        #
        self.d_t = self.tline.step
        F_ext = np.zeros_like(self.U_n)
        t_n = self.tline.val
        t_n1 = t_n
        U_n = self.U_n
        while (t_n1 - self.tline.max) <= self.step_tolerance and \
                not (self.restart or self.paused):
            k = 0
            print('load factor', t_n1, end=' ')
            step_flag = 'predictor'
            U_k = np.copy(U_n)
            d_U_k = np.zeros_like(U_k)
            while k <= self.k_max and \
                    not (self.restart or self.paused):

                self.K.reset_mtx()

                K_mtx, F_int = self.ts.get_corr_pred(
                    U_k, d_U_k, t_n, t_n1,
                    step_flag,
                    **self.state_arrays
                )

                self.K.add_mtx(K_mtx)

                # Prepare F_ext by zeroing it
                #
                F_ext[:] = 0.0

                # Assemble boundary conditions in K and self.F_ext
                #
                self.ts.bcond_mngr.apply(
                    step_flag, None, self.K, F_ext, t_n, t_n1)

                # Return the system matrix assembly K and the residuum
                #
                R = F_ext - F_int

                self.K.apply_constraints(R)
                d_U_k, pos_def = self.K.solve()
                U_k += d_U_k
                if np.linalg.norm(R) < self.tolerance:
                    U_n[:] = U_k[:]
                    self.t_record.append(t_n1)
                    self.U_record.append(U_k)
                    self.F_record.append(F_int)
                    self.state_record.append(np.copy(self.state))
                    break
                k += 1
                step_flag = 'corrector'

            if k >= self.k_max:
                print(' ----------> no convergence')
                break
            else:
                print('(', k, ')')
            if self.restart or self.paused:
                print('interrupted iteration')
                break

            t_n = t_n1
            t_n1 = t_n + self.d_t
            self.tline.val = min(t_n, self.tline.max)
        return U_k

    def get_time_idx_arr(self, vot):
        '''Get the index corresponding to visual time
        '''
        x = self.t_record
        idx = np.array(np.arange(len(x)), dtype=np.float64)
        t_idx = np.interp(vot, x, idx)
        return np.array(t_idx + 0.5, np.int_)

    def get_time_idx(self, vot):
        return int(self.get_time_idx_arr(vot))


if __name__ == '__main__':

    from ibvpy.api import BCDof
    from ibvpy.tmodel.mats3D.mats3D_elastic.mats3D_elastic import \
        MATS3DElastic
    from ibvpy.tmodel.mats3D.mats3D_plastic.mats3D_desmorat import \
        MATS3DDesmorat
    from ibvpy.tmodel.mats3D import \
        MATS3DMplCSDEEQ
    ts = MATS3DExplore(
        mats_eval=MATS3DDesmorat()
    )

    ts.bcond_mngr.bcond_list = [BCDof(var='u', dof=0, value=0.001)]

    tl = TLoop(ts=ts, tline=TLine(step=0.1))
    tl.init()
    U_k = tl.eval()

    print(U_k)

    print('U', tl.U_record)
    print('F', tl.F_record)




from traits.api import Float, Callable, on_trait_change
from bmcs_utils.api import View, Item
from ibvpy.view.ui import BMCSLeafNode
import bmcs_utils.api as bu


class TLine(BMCSLeafNode):
    '''
    Time line for the control parameter.

    This class sets the time-range of the computation - the start and stop time.
    val is the value of the current time.

    TODO - the info page including the number of load steps
    and estimated computation time.

    TODO - the slide bar is not read-only. How to include a real progress bar?
    '''
    node_name = 'time range'

    min = bu.Float(0.0,
                TIME=True
                )
    max = bu.Float(1.0,
                TIME=True
                )
    step = bu.Float(0.1,
                 TIME=True
                 )
    val = bu.Float(0.0)

    def _val_changed(self):
        if self.time_change_notifier:
            self.time_change_notifier(self.val)

    @on_trait_change('min,max')
    def _time_range_changed(self):
        if self.time_range_change_notifier:
            self.time_range_change_notifier(self.max)

    time_change_notifier = Callable

    time_range_change_notifier = Callable

    ipw_view = View(
        Item('min', full_size=True),
        Item('max'),
        Item('step'),
        Item('val', style='readonly')
    )




from .loading_scenario import \
    LoadingScenario, Viz2DLoadControlFunction, \
    MonotonicLoadingScenario, CyclicLoadingScenario

from .time_function import TimeFunction, TFSelector, \
    TFMonotonic, TFCyclicSymmetricIncreasing, TFCyclicSymmetricConstant, \
    TFCyclicNonsymmetricIncreasing, TFCyclicNonsymmetricConstant, \
    TFCyclicSin, TFBilinear



from .mats2D_elastic.vmats2D_elastic import MATS2DElastic
from .mats2D_microplane.vmats2D_mpl_csd_eeq import MATS2DMplCSDEEQ
from .mats2D_microplane.vmats2D_mpl_d_eeq import MATS2DMplDamageEEQ
from .mats2D_sdamage.vmats2D_sdamage import MATS2DScalarDamage


#from mats2D_microplane.vmats2D_mpl_d_odf import MATS2DMplDamageODF



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Jun 30, 2009 by: rchx


from traits.api import Interface, Int, Array, Property


class IFENodeSlice(Interface):
    '''
    Interface of the spatial domain.
    '''

    geo_X = Property

    dof_X = Property

    dofs = Property


class IFEGridSlice(Interface):
    '''
    Interface of the spatial domain.
    '''

    elems = Property(Array(Int))

    geo_X = Property

    dof_X = Property

    dofs = Property



'''
Created on Feb 8, 2018

@author: rch
'''

from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
import numpy as np


class MATS2DElastic(MATS2DEval):
    '''Elastic material model.
    '''
    state_var_shapes = {}

    def get_corr_pred(self, eps_Emcd, t_n1):
        D_Emabcd = self.D_abcd[None, None, :, :, :, :]
        sig_Emab = np.einsum('abcd,...cd->...ab', self.D_abcd, eps_Emcd)
        return sig_Emab, D_Emabcd



#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Jul 19, 2010 by: rch

from ibvpy.mesh.cell_grid.cell_grid import CellGrid
from ibvpy.mesh.cell_grid.cell_spec import CellSpec

# Get the intersected element of a one dimensional grid
cell_grid = CellGrid( grid_cell_spec = CellSpec( node_coords = [[-1, -1],
                                                                [-1, 0],
                                                                [-1, 1],
                                                                [ 1, -1],
                                                                [ 1, 0],
                                                                [ 1, 1]]
                                                                ),
                      shape = ( 1, 1 ), coord_max = [ 1., 1. ] )

print('vertex_idx_grid')
print(cell_grid.vertex_idx_grid)
print('vertex slices')
print(cell_grid.vertex_slices)
print('point_x_grid')
print(cell_grid.point_x_grid)
print('point_x_arr')
print(cell_grid.point_x_arr)
print('point_X_grid')
print(cell_grid.point_X_grid)
print('vertex_X_grid', end=' ')
print(cell_grid.vertex_X_grid)




#-------------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Jan 21, 2011 by: rch


from ibvpy.tmodel.mats2D.mats2D_elastic.mats2D_elastic import MATS2DElastic
from ibvpy.tmodel.mats1D.mats1D_elastic.mats1D_elastic import MATS1DElastic
from ibvpy.api import FEDomain, FEGrid, FERefinementGrid, TStepper as TS, TLoop, \
    BCSlice, RTraceDomainListField
from ibvpy.fets.fets2D.fets2D4q import FETS2D4Q
from ibvpy.fets.fets1D.fets1D2l import FETS1D2L

if __name__ == '__main__':

    if True:
        fets_eval_4u = FETS2D4Q( mats_eval = MATS2DElastic() )
        fe_grid = FEGrid( name = 'fe_grid1', coord_max = ( 4., 4. ),
                                   shape = ( 4, 4 ),
                                   fets_eval = fets_eval_4u )

        interior_elems = fe_grid[ 1:3, 1:3, :, : ].elems
        interior_bc = fe_grid[ 1, 1, 1:, 1: ]

        bcond_list = [BCSlice( var = 'u', dims = [0, 1], slice = fe_grid[ :, 0, :, 0 ], value = 0.0 ),
                      BCSlice( var = 'u', dims = [0, 1], slice = interior_bc,
                               link_slice = fe_grid[1, 0, 0, 0], link_coeffs = [0], value = 0.0 ),
                      BCSlice( var = 'f', dims = [1], slice = fe_grid[ 0, -1, :, -1 ], value = 1.0 ) ]

    else:
        fets_eval_1d = FETS1D2L( mats_eval = MATS1DElastic() )
        fe_grid = FEGrid( name = 'fe_grid1', coord_max = ( 4., ),
                                   shape = ( 4, ),
                                   fets_eval = fets_eval_1d )

        interior_elems = fe_grid[ 1:3, : ].elems
        interior_bc = fe_grid[ 1:2, 1: ]

        bcond_list = [BCSlice( var = 'u', dims = [0], slice = fe_grid[ 0, 0 ], value = 0.0 ),
                      BCSlice( var = 'u', dims = [0], slice = interior_bc,
                               link_slice = fe_grid[0, 0], link_coeffs = [0], value = 0.0 ),
                      BCSlice( var = 'u', dims = [0], slice = fe_grid[ -1, -1 ], value = 1.0 ) ]

    fe_grid.inactive_elems = list( interior_elems )
    print('elems', interior_bc.elems)
    print('dofs', interior_bc.dofs)
    print('nodes', interior_bc.dof_nodes)

    rtrace_list = [RTraceDomainListField( name = 'sig_app' ,
                                      var = 'sig_app',
                                      record_on = 'update', ) ]

    ts = TS( sdomain = fe_grid,
             bcond_list = bcond_list,
             rtrace_list = rtrace_list,
             )
    tloop = TLoop( tstepper = ts,
                    )

    u = tloop.eval()

#    from ibvpy.plugins.ibvpy_app import IBVPyApp
#    ibvpy_app = IBVPyApp( ibv_resource = tloop )
#    ibvpy_app.main()




from ibvpy.tmodel.mats2D.mats2D_eval import MATS2DEval
from bmcs_utils.api import Float, Item, View, Enum, EitherType, FloatRangeEditor
import numpy as np
import traits.api as tr
from ibvpy.tmodel.mats_damage_fn import \
    IDamageFn, GfDamageFn, ExpSlopeDamageFn, AbaqusDamageFn, \
    LinearDamageFn, FRPDamageFn, WeibullDamageFn

from .vstrain_norm2d import StrainNorm2D, SN2DRankine, SN2DMasars, SN2DEnergy

class MATS2DScalarDamage(MATS2DEval):
    r'''Isotropic damage model.
    '''

    name = 'isotropic damage model'
    node_name = 'isotropic damage model'

    depends_on = ['omega_fn','strain_norm']
    tree = ['omega_fn','strain_norm']

    omega_fn = EitherType(
        options=[('exp-slope', ExpSlopeDamageFn),
                 ('linear', LinearDamageFn),
                 ('abaqus', AbaqusDamageFn),
                 ('fracture-energy', GfDamageFn),
                 ('weibull-CDF', WeibullDamageFn),
                 ],
        MAT=True,
        on_option_change='link_omega_to_mats'
    )

    # upon change of the type attribute set the link to the material model
    def link_omega_to_mats(self):
        self.omega_fn_.trait_set(mats=self,
                                 E_name='E',
                                 x_max_name='eps_max')

    D_alg = Float(0)
    r'''Selector of the stiffness calculation.
    '''

    #=========================================================================
    # Material model
    #=========================================================================
    strain_norm = EitherType(
        options=[('Rankine', SN2DRankine),
                 ('Masars', SN2DMasars),
                 ('Energy', SN2DEnergy)],
        on_option_change='link_strain_norm_to_mats'
    )

    def link_strain_norm_to_mats(self):
        self.strain_norm_.trait_set(mats=self)

    state_var_shapes = {'kappa': (),
                        'omega': ()}
    r'''
    Shapes of the state variables
    to be stored in the global array at the level 
    of the domain.
    '''

    def get_corr_pred(self, eps_ab_n1, tn1, kappa, omega):
        r'''
        Corrector predictor computation.
        @param eps_app_eng input variable - engineering strain
        '''
        eps_eq = self.strain_norm_.get_eps_eq(eps_ab_n1, kappa)
        I = self.omega_fn_.get_f_trial(eps_eq, kappa)
        eps_eq_I = eps_eq[I]
        kappa[I] = eps_eq_I
        omega[I] = self._omega(eps_eq_I)
        phi = (1.0 - omega)
        D_abcd = np.einsum(
            '...,abcd->...abcd',
            phi, self.D_abcd
        )
        sig_ab = np.einsum(
            '...abcd,...cd->...ab',
            D_abcd, eps_ab_n1
        )
        if self.D_alg > 0:
            domega_ds_I = self._omega_derivative(eps_eq_I)
            deps_eq_I = self.strain_norm_.get_deps_eq(eps_ab_n1[I])
            D_red_I = np.einsum('...,...ef,cdef,...ef->...cdef', domega_ds_I,
                                deps_eq_I, self.D_abcd, eps_ab_n1[I]) * self.D_alg
            D_abcd[I] -= D_red_I

        return sig_ab, D_abcd

    def _omega(self, kappa):
        return self.omega_fn_(kappa)

    def _omega_derivative(self, kappa):
        return self.omega_fn_.diff(kappa)

    ipw_view = View(
        *MATS2DEval.ipw_view.content,
        Item('strain_norm'),
        Item('omega_fn'),
        Item('stress_state'),
        Item('D_alg', latex=r'\theta_\mathrm{alg. stiff.}',
                editor=FloatRangeEditor(low=0,high=1)),
    )

    def get_omega(self, eps_ab, tn1, **Eps):
        return Eps['omega']

    var_dict = tr.Property(tr.Dict(tr.Str, tr.Callable))
    '''Dictionary of response variables
    '''
    @tr.cached_property
    def _get_var_dict(self):
        var_dict = dict(omega=self.get_omega)
        var_dict.update(super()._get_var_dict())
        return var_dict





from traits.api import Interface, List, Int, Array, Property

class IFEParentDomain( Interface ):
    '''
    Interface of the spatial domain.
    '''

        
    n_active_elems = Property
    
    def deactivate(self, idx_tuple ):
        '''Deactivate an element with the tuple index idx.
        '''
    
    def reactivate(self, idx_tuple ):
        '''Reactivate an element that has previously been deactivated.
        '''
    




import mayavi.mlab as m
import numpy as np
import bmcs_utils.api as bu
from bmcs_utils.api import Float, Int, Model
ONE = np.ones((1,), dtype=np.float64)
DELTA = np.identity(3)

# Levi Civita symbol
EPS = np.zeros((3, 3, 3), dtype='f')
EPS[(0, 1, 2), (1, 2, 0), (2, 0, 1)] = 1
EPS[(2, 1, 0), (1, 0, 2), (0, 2, 1)] = -1


class YieldConditionJ2(Model):
    '''J2-plasticity
    '''
    n_D = Int(3)
    sig_y = Float(2.0)
    H = Float(10.0)

    def f(self, sig_ij, z=0):
        '''Given a tensor sig_ij return the yield condition
        '''
        I1 = np.einsum('...ii,...ii', sig_ij, DELTA)
        s_ij = sig_ij - np.einsum('...,ij->...ij', I1 / 3.0, DELTA)
        J2 = np.einsum('...ii,...ii', s_ij, s_ij) / 2.0

        return J2 - self.sig_y - self.H * z

    def get_df_dsig(self):
        pass


class YieldConditionDruckerPrager(Model):
    '''Drucker-Prager yield condition
    '''
    n_D = Int(3)
    alpha_F = Float(0.24)
    tau2_y = Float(2.6)
#     f_t = Float(3.0)
#     f_c = Float(30.0)

    def f(self, sig_ij):
        '''Given a tensor sig_ij return the yield condition
        '''
        I1 = np.einsum('...ii,...ii', sig_ij, DELTA)
        s_ij = sig_ij - np.einsum('...,ij->...ij', I1 / 3.0, DELTA)
        J2 = np.einsum('...ij,...ij', s_ij, s_ij) / 2.0

        #alpha_F = (self.f_c - self.f_t) / 3.0
        alpha_F = self.alpha_F
        tau2_y = self.tau2_y
        #tau2_y = self.f_c * self.f_t / 3.0

        return np.sqrt(J2) + alpha_F * I1 - tau2_y

    def get_df_dsig(self):
        pass

    ipw_view = bu.View(
        bu.Item('alpha_F'),
        bu.Item('tau2_y'))


class YieldConditionVonMises(HasStrictTraits):

    '''
    '''
    k = Float(3)

    def f(self, sig_ij):
        I1 = np.einsum('...ii,...ii', sig_ij, DELTA)
        s_ij = sig_ij - np.einsum('...,ij->...ij', I1 / 3.0, DELTA)
        J2 = np.einsum('...ij,...ij', s_ij, s_ij) / 2.0

        return J2 - self.k ** 2

    view = bu.View(
        bu.Item('k'),
    )


class YieldConditionWillamWarnke(Model):

    '''the three parameter Willam-Warnke yield function,
    https://en.wikipedia.org/wiki/Willam-Warnke_yield_criterion
    '''
    sig_c = Float(10.)  # the uniaxial compressive strength
    sig_t = Float(3.)  # the uniaxial tensile strength
    sig_b = Float(17.)  # the equibiaxial compressive strength

    def f(self, sig_ij):
        I1 = np.einsum('...ii,...ii', sig_ij, DELTA)
        s_ij = sig_ij - np.einsum('...,ij->...ij', I1 / 3.0, DELTA)
        J2 = np.einsum('...ij,...ij', s_ij, s_ij) / 2.0
        J3 = np.einsum('...ij,...jk,...ki', s_ij, s_ij, s_ij) / 3.0
        # the Lode angle

        theta = np.arccos(1.5 * np.sqrt(3.0) * J3 / J2 ** (1.5)) / 3.0

        rc = np.sqrt(1.2) * self.sig_b * self.sig_t / \
            (3 * self.sig_b * self.sig_t +
             self.sig_c * (self.sig_b - self.sig_t))
        rt = np.sqrt(1.2) * self.sig_b * self.sig_t / \
            (self.sig_c * (2 * self.sig_b + self.sig_t))

        u = 2 * rc * (rc ** 2 - rt ** 2) * np.cos(theta)

        a = 4 * (rc ** 2 - rt ** 2) * np.cos(theta) ** 2 + \
            5 * rt ** 2 - 4 * rt * rc
        v = rc * (2 * rt - rc) * np.sqrt(a)

        w = 4 * (rc ** 2 - rt ** 2) * np.cos(theta) ** 2 + (rc - 2 * rt) ** 2

        r = (u + v) / w
        z = self.sig_b * self.sig_t / self.sig_c / (self.sig_b - self.sig_t)

        return (1. / (3. * z) * I1 / self.sig_c + np.sqrt(0.4) /
                r * np.sqrt(J2) / self.sig_c - 1.)

    view = bu.View(
        bu.Item('sig_c'),
        bu.Item('sig_t'),
        bu.Item('sig_b'))


class YieldConditionRankine(Model):

    '''single-parameter Rankine failure surface
    '''
    sig_t = Float(3.)  # the uniaxial tensile strength

    def f(self, sig_ij):
        sig_evals, sig_evects = np.linalg.eigh(sig_ij)
        max_sig = np.max(sig_evals, axis=-1)
        f = max_sig - self.sig_t
        return f

    view = bu.View(
        bu.Item('sig_t'),
    )


class YieldConditionAbaqus(Model):

    '''Lee et al (1988) - PLASTIC-DAMAGE MODEL FOR CYCLIC
     LOADING OF CONCRETE STRUCTURES'''

    sig_c = Float(-10.)  # the uniaxial compressive strength
    sig_t = Float(3.)  # the uniaxial tensile strength
    sig_b = Float(-17.)  # the equibiaxial compressive strength

    def f(self, sig_ij):
        I1 = np.einsum('...ii,...ii', sig_ij, DELTA)
        s_ij = sig_ij - np.einsum('...,ij->...ij', I1 / 3.0, DELTA)
        J2 = np.einsum('...ij,...ij', s_ij, s_ij) / 2.0

        alpha = (self.sig_b - self.sig_c) / (2. * self.sig_b - self.sig_c)
        beta = self.sig_c / self.sig_t * (alpha - 1.) - (1. + alpha)

        sig_i = np.einsum('...ii->...i', sig_ij)
        sig_max = np.amax(sig_i, axis=-1)

        F = 1. / (1. - alpha) * \
            (alpha * I1 + np.sqrt(3. * J2) + beta * sig_max)
        c = -self.sig_c  # cohesion

        return F - c


view = bu.View(bu.Item('sig_c'),
               bu.Item('sig_t'),
               bu.Item('sig_b'))


class YieldConditionExtendedLeonModel(Model):
    """
    """

    sig_c = Float(-10.)  # the uniaxial compressive strength
    sig_t = Float(3.)  # the uniaxial tensile strength
    sig_b = Float(-17.)  # the equibiaxial compressive strength

    def f(self, sig_ij):
        I1 = np.einsum('...ii,...ii', sig_ij, DELTA)
        s_ij = sig_ij - np.einsum('...,ij->...ij', I1 / 3.0, DELTA)
        J2 = np.einsum('...ij,...ij', s_ij, s_ij) / 2.0
        J3 = np.einsum('...ij,...jk,...ki', s_ij, s_ij, s_ij) / 3.0
        z = I1 / np.sqrt(3.0)
        r = np.sqrt(2.0 * J2)
        sin3theta = J3 / 2.0 * np.power(3.0 / J2, 1.5)
        theta = np.arcsin(sin3theta) / 3.0

        # -- here to go on

        return z, r, theta

    view = bu.View(
        bu.Item('sig_c'),
                bu.Item('sig_t'),
                bu.Item('sig_b'))


def get_lut():
    opacity = 20.0
    lut = np.zeros((256, 4), dtype=Int)
    alpha = 255 * opacity / 100.0
    lut[:] = np.array([0, 0, 255, int(round(alpha))], dtype=Int)


def show_yield_faces():
    #     yc = YieldConditionJ2(sig_y=6.0)
    sig = np.array([[2, 3, 4],
                    [1, 3, 2],
                    [3, 4, 5]], dtype=np.float64)

    yc = YieldConditionExtendedLeonModel()
    f = yc.f(sig)
    print('f', f)

    # return

    min_sig = -20.0
    max_sig = 5.0
    n_sig = 100j
    sig_1, sig_2, sig_3 = np.mgrid[min_sig: max_sig: n_sig,
                                   min_sig: max_sig: n_sig,
                                   min_sig: max_sig: n_sig]

    sig_abcj = np.einsum('jabc->abcj', np.array([sig_1, sig_2, sig_3]))
    sig_abcij = np.einsum('abcj,jl->abcjl', sig_abcj, DELTA)

    yc = YieldConditionDruckerPrager(alpha_F=0.24, tau2_y=2.6)
    f = yc.f(sig_abcij)
    f_pipe = m.contour3d(
        sig_1, sig_2, sig_3, f, contours=[0.0], color=(1, 0, 0))
    f_pipe.module_manager.scalar_lut_manager.lut.table = get_lut()

#     yc = YieldConditionVonMises(k=10.)
    yc = YieldConditionWillamWarnke()
    f = yc.f(sig_abcij)
    f_pipe = m.contour3d(
        sig_1, sig_2, sig_3, f, contours=[0.0], color=(0, 1, 0))
    f_pipe.module_manager.scalar_lut_manager.lut.table = get_lut()

    yc = YieldConditionAbaqus()
    f = yc.f(sig_abcij)
    f_pipe = m.contour3d(
        sig_1, sig_2, sig_3, f, contours=[0.0], color=(0, 0, 1))
    f_pipe.module_manager.scalar_lut_manager.lut.table = get_lut()

    m.axes(f_pipe)
    m.show()


if __name__ == '__main__':
    show_yield_faces()



from traits.api import \
    Bool, Enum, HasTraits, \
    Instance, Trait, Str, \
    Callable, Tuple

from numpy import \
    array

from tvtk.api import tvtk
# Mayavi related imports
#
from mayavi.sources.api import VTKDataSource, VTKFileReader
from mayavi.modules.api import Surface
from mayavi.filters.api import ExtractTensorComponents, \
    ExtractVectorComponents

from mayavi.filters.api import WarpVector

# MayaVI engine used for the pipeline construction
#
from .mayavi_engine import get_engine


class MVStructuredGrid(HasTraits):
    dims = Callable

    def _dims(self):
        return lambda: (1, 1, 1)

    points = Callable

    def _points_default(self):
        return lambda: array([])

    pd = Instance(tvtk.StructuredGrid)

    def _pd_default(self):
        return tvtk.StructuredGrid()

    scalars = Callable

    vectors = Callable

    tensors = Callable

    name = Str('')

    def __init__(self, **kw):

        super(MVStructuredGrid, self).__init__(**kw)
        e = get_engine()

        from mayavi.modules.api import \
            StructuredGridOutline, GridPlane

        self.src = VTKDataSource(name=self.name, data=self.pd)
        e.add_source(self.src)

        o = StructuredGridOutline()
        e.add_module(o)

        for axis in ['x', 'y', 'z']:
            g = GridPlane(name='%s - grid plane' % axis)
            g.grid_plane.axis = axis
            e.add_module(g)

        if self.scalars or self.vectors or self.tensors:
            s = Surface()
            e.add_module(s)

    def redraw(self):

        self.pd.dimensions = self.dims()
        self.pd.points = self.points()

        if self.scalars:
            self.pd.point_data.scalars = self.scalars()
        if self.vectors:
            self.pd.point_data.vectors = self.vectors()
        if self.tensors:
            self.pd.point_data.tensors = self.tensors()

        self.src.data_changed = True


class MVUnstructuredGrid(HasTraits):

    #    vtk_cell_array = Instance(tvtk.CellArray())
    #    def _vtk_cell_array_default(self):
    #        return tvtk.CellArray()

    warp = Bool(False)

    name = Str
    position = Enum('nodes', 'int_pnts')
    engine = Trait()

    def _engine_default(self):
        return get_engine()

    def rebuild_pipeline(self, pd):
        src = VTKDataSource(name=self.name, data=pd)
        self.engine.add_source(src)
        if self.warp:
            self.engine.add_filter(WarpVector())

        if self.name in src._point_tensors_list:
            src.point_tensors_name = self.name
            self.engine.add_filter(ExtractTensorComponents())
        elif self.name in src._point_vectors_list:
            src.point_vectors_name = self.name
            self.engine.add_filter(ExtractVectorComponents())
        elif self.name in src._point_scalars_list:
            src.point_scalars_name = self.name
        s = Surface()
        s.actor.property.point_size = 5.
        self.engine.add_module(s)
        src.scene.z_plus_view()


class MVVTKSource(HasTraits):

    warp = Bool(False)

    name = Str
    position = Enum('nodes', 'int_pnts')
    engine = Trait()

    def _engine_default(self):
        return get_engine()

#    def redraw(self):
#        e = get_engine()
# directory management
#        home_dir = os.environ['HOME']
#        base_name = os.path.basename(os.getcwd())
#
#        data_dir = home_dir+'/simdata'
#        if not os.path.exists(data_dir):
#            os.mkdir(data_dir)
#            print "simdata directory created"
#
#        if not os.path.exists(data_dir +'/'+ base_name):
#            os.mkdir(data_dir +'/'+ base_name)
#            print base_name," directory created"
#
#        os.chdir(data_dir +'/'+ base_name)
#
#        self.src = VTKFileReader(base_file_name = 'nodes_0.vtk')
# self.src = VTKDataSource( name = self.name, data = self.pd )
# self.e.add_source(self.src)
#        e.add_source(self.src)
#        g = Surface()
#        e.add_module(g)
# g.module_manager.scalar_lut_manager.show_scalar_bar = True # show scalar bar
#
#        self.src.scene.z_plus_view()
# self.src.data_changed = True

    def rebuild_pipeline(self, pd):
        src = VTKFileReader(base_file_name='%s_0.vtk' % self.position)
        self.engine.add_source(src)
        if self.warp:
            self.engine.add_filter(WarpVector())

        if self.name in src._point_tensors_list:
            src.point_tensors_name = self.name
            self.engine.add_filter(ExtractTensorComponents())
        elif self.name in src._point_vectors_list:
            src.point_vectors_name = self.name
            self.engine.add_filter(ExtractVectorComponents())
        elif self.name in src._point_scalars_list:
            src.point_scalars_name = self.name
        s = Surface()
        s.actor.property.point_size = 5.
        self.engine.add_module(s)
        src.scene.z_plus_view()


class MVPBase(HasTraits):

    '''
    Mayavi Pipeline Base class
    '''

    points = Callable

    def _points_default(self):
        return lambda: array([])

    lines = Callable

    def _lines_default(self):
        return lambda: array([])

    polys = Callable

    def _polys_default(self):
        return lambda: array([])

    scalars = Callable

    vectors = Callable

    tensors = Callable

    pd = Instance(tvtk.PolyData)

    def _pd_default(self):
        return tvtk.PolyData()

    def redraw(self):

        self.pd.points = self.points()
        self.pd.lines = self.lines()
        self.pd.polys = self.polys()
        if self.scalars:
            self.pd.point_data.scalars = self.scalars()
        if self.vectors:
            self.pd.point_data.vectors = self.vectors()
        if self.tensors:
            self.pd.point_data.tensors = self.tensors()
        self.src.data_changed = True


#-------------------------------------------------------------------
# MVMeshSource - mayavi mesh source
#-------------------------------------------------------------------

class MVPolyData(MVPBase):

    '''
    Provide a Mayavi source for polar visualization visualization. 

    @TODO must be revised in the context of the view and editor
    concept.  It is now a mixin-base class providing the visualization
    functionality. Still its the question whether or not it should be
    implemented taht way.
    '''

    name = Str('')

    def __init__(self, **kw):

        e = get_engine()

        super(MVPolyData, self).__init__(**kw)
        from mayavi.modules.api import Outline, Surface, Labels

        self.src = VTKDataSource(name=self.name, data=self.pd)
        e.add_source(self.src)

        o = Outline()
        e.add_module(o)
        s = Surface()
        e.add_module(s)


class MVPointLabels(MVPBase):
    color = Tuple(1., 1., 1.)

    def __init__(self, **kw):

        e = get_engine()

        super(MVPointLabels, self).__init__(**kw)
        from mayavi.modules.api import Outline, Surface, Labels

        self.src = VTKDataSource(name=self.name, data=self.pd)
        e.add_source(self.src)

        self.labels = Labels(name='Node numbers', object=self.src,
                             label_format='%g',
                             number_of_labels=100)
        self.labels.property.color = self.color
        e.add_module(self.labels)

    def redraw(self, label_mode='label_ids'):
        super(MVPointLabels, self).redraw()
        self.labels.mapper.label_mode = label_mode



from numpy import \
    zeros, dot, hstack, identity
from scipy.linalg import \
    inv
from traits.api import \
    Instance, Int, Property, Array, cached_property

from .fets3D import FETS3D


#-------------------------------------------------------------------------
# FETS3D8H - 8 nodes isoparametric volume element (3D, linear, Lagrange family)
#-------------------------------------------------------------------------
class FETS3D8H(FETS3D):
    '''
    eight nodes volume element
    '''
    debug_on = False

    # number of nodal degrees of freedom
    # number of degrees of freedom of each element
    n_nodal_dofs = Int(3)
    n_e_dofs = Int(8 * 3)

    # Integration parameters
    #
    ngp_r = 2
    ngp_s = 2
    ngp_t = 2

    # Here an isoparametric element formulation is applied.
    # The implemented shape functions are derived based on the following
    # ordering of the nodes of the parent element

    dof_r = \
        Array(value=[[-1., -1., -1.],
                     [1., -1., -1.],
                     [-1., 1., -1.],
                     [1., 1., -1.],
                     [-1., -1., 1.],
                     [1., -1., 1.],
                     [-1., 1., 1.],
                     [1., 1., 1.]])

    geo_r = \
        Array(value=[[-1., -1., -1.],
                     [1., -1., -1.],
                     [-1., 1., -1.],
                     [1., 1., -1.],
                     [-1., -1., 1.],
                     [1., -1., 1.],
                     [-1., 1., 1.],
                     [1., 1., 1.]])

    # Used for Visualization
    vtk_r = Array(value=[[-1., -1., -1.],
                         [1., -1., -1.],
                         [-1., 1., -1.],
                         [1., 1., -1.],
                         [-1., -1., 1.],
                         [1., -1., 1.],
                         [-1., 1., 1.],
                         [1., 1., 1.]])
    vtk_cells = [[0, 1, 3, 2, 4, 5, 7, 6]]
    vtk_cell_types = 'Hexahedron'

    #---------------------------------------------------------------------
    # Method required to represent the element geometry
    #---------------------------------------------------------------------
    def get_N_geo_mtx(self, r_pnt):
        '''
        Return the value of shape functions (derived in femple) for the
        specified local coordinate r_pnt
        '''
        N_geo_mtx = zeros((1, 8), dtype='float64')
        N_geo_mtx[0, 0] = -((-1 + r_pnt[2]) * (-1 + r_pnt[1]) *
                            (-1 + r_pnt[0])) / 8.0
        N_geo_mtx[0, 1] = ((-1 + r_pnt[2]) * (-1 + r_pnt[1]) *
                           (1 + r_pnt[0])) / 8.0
        N_geo_mtx[0, 2] = ((-1 + r_pnt[2]) * (1 + r_pnt[1]) *
                           (-1 + r_pnt[0])) / 8.0
        N_geo_mtx[0, 3] = -((-1 + r_pnt[2]) * (1 + r_pnt[1]) *
                            (1 + r_pnt[0])) / 8.0
        N_geo_mtx[0, 4] = ((1 + r_pnt[2]) * (-1 + r_pnt[1]) *
                           (-1 + r_pnt[0])) / 8.0
        N_geo_mtx[0, 5] = -((1 + r_pnt[2]) * (-1 + r_pnt[1]) *
                            (1 + r_pnt[0])) / 8.0
        N_geo_mtx[0, 6] = -((1 + r_pnt[2]) * (1 + r_pnt[1]) *
                            (-1 + r_pnt[0])) / 8.0
        N_geo_mtx[0, 7] = ((1 + r_pnt[2]) * (1 + r_pnt[1]) *
                           (1 + r_pnt[0])) / 8.0
        return N_geo_mtx

    def get_dNr_geo_mtx(self, r_pnt):
        '''
        Return the matrix of shape function derivatives (derived in femple).
        Used for the construction of the Jacobi matrix.
        '''
        dNr_geo_mtx = zeros((3, 8), dtype='float64')
        dNr_geo_mtx[0, 0] = -((-1 + r_pnt[2]) * (-1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[0, 1] = ((-1 + r_pnt[2]) * (-1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[0, 2] = ((-1 + r_pnt[2]) * (1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[0, 3] = -((-1 + r_pnt[2]) * (1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[0, 4] = ((1 + r_pnt[2]) * (-1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[0, 5] = -((1 + r_pnt[2]) * (-1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[0, 6] = -((1 + r_pnt[2]) * (1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[0, 7] = ((1 + r_pnt[2]) * (1 + r_pnt[1])) / 8.0
        dNr_geo_mtx[1, 0] = -((-1 + r_pnt[2]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[1, 1] = ((-1 + r_pnt[2]) * (1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[1, 2] = ((-1 + r_pnt[2]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[1, 3] = -((-1 + r_pnt[2]) * (1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[1, 4] = ((1 + r_pnt[2]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[1, 5] = -((1 + r_pnt[2]) * (1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[1, 6] = -((1 + r_pnt[2]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[1, 7] = ((1 + r_pnt[2]) * (1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 0] = -((-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 1] = ((-1 + r_pnt[1]) * (1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 2] = ((1 + r_pnt[1]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 3] = -((1 + r_pnt[1]) * (1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 4] = ((-1 + r_pnt[1]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 5] = -((-1 + r_pnt[1]) * (1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 6] = -((1 + r_pnt[1]) * (-1 + r_pnt[0])) / 8.0
        dNr_geo_mtx[2, 7] = ((1 + r_pnt[1]) * (1 + r_pnt[0])) / 8.0
        return dNr_geo_mtx

    #---------------------------------------------------------------------
    # Method delivering the shape functions for the field variables and their
    # derivatives
    #---------------------------------------------------------------------
    def get_N_mtx(self, r_pnt):
        '''
        Returns the matrix of the shape functions used for the field approximation
        containing zero entries. The number of rows corresponds to the number of nodal
        dofs. The matrix is evaluated for the specified local coordinate r_pnt.
        '''
        N = self.get_N_geo_mtx(r_pnt)
        I_mtx = identity(self.n_nodal_dofs, float)
        N_mtx_list = [I_mtx * N[0, i] for i in range(0, N.shape[1])]
        N_mtx = hstack(N_mtx_list)
        return N_mtx

    def get_dNr_mtx(self, r_pnt):
        '''
        Return the derivatives of the shape functions used for the field approximation
        '''
        dNr_mtx = self.get_dNr_geo_mtx(r_pnt)
        return dNr_mtx

#----------------------- example --------------------


if __name__ == '__main__':
    from ibvpy.api import \
        TStepper as TS, RTDofGraph, RTraceDomainListField, TLoop, \
        TLine, BCDofGroup, IBVPSolve as IS, DOTSEval

    from ibvpy.tmodel.mats3D.mats3D_elastic.mats3D_elastic import MATS3DElastic

    fets_eval = FETS3D8H(mats_eval=MATS3DElastic(nu=0.25))

    from ibvpy.mesh.fe_grid import FEGrid

    # Discretization
    domain = FEGrid(coord_max=(3., 3., 3.),
                    shape=(10, 10, 10),
                    fets_eval=fets_eval)

    ts = TS(
        sdomain=domain,
        bcond_list=[BCDofGroup(var='u', value=0., dims=[0],
                               get_dof_method=domain.get_left_dofs),
                    BCDofGroup(var='u', value=0., dims=[1, 2],
                               get_dof_method=domain.get_bottom_left_dofs),
                    BCDofGroup(var='u', value=0.002, dims=[0],
                               get_dof_method=domain.get_right_dofs)],
        rtrace_list=[
            #                        RTDofGraph(name = 'Fi,right over u_right (iteration)' ,
            #                                  var_y = 'F_int', idx_y = right_dof,
            #                                  var_x = 'U_k', idx_x = right_dof,
            #                                  record_on = 'update'),
            RTraceDomainListField(name='Deformation',
                                       var='eps_app', idx=0,
                                       record_on='update'),
            RTraceDomainListField(name='Displacement',
                                       var='u', idx=0),
            #                         RTraceDomainListField(name = 'Stress' ,
            #                                        var = 'sig', idx = 0,
            #                                        record_on = 'update'),
            #                        RTraceDomainListField(name = 'N0' ,
            #                                       var = 'N_mtx', idx = 0,
            #                                       record_on = 'update')
        ]
    )

    # Add the time-loop control
    tloop = TLoop(tstepper=ts,
                  tline=TLine(min=0.0, step=0.5, max=1.0))

    tloop.eval()

    # Put the whole stuff into the simulation-framework to map the
    # individual pieces of definition into the user interface.
    #
    from ibvpy.plugins.ibvpy_app import IBVPyApp
    app = IBVPyApp(ibv_resource=tloop)
    app.main()



#-------------------------------------------------------------------------
#
# Copyright (c) 2009, IMB, RWTH Aachen.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in simvisage/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.simvisage.com/licenses/BSD.txt
#
# Thanks for using Simvisage open source!
#
# Created on Aug 19, 2009 by: rch

from ibvpy.tmodel.mats3D.__test__ import TestMATS3D

from .vmats3D_elastic import \
    MATS3DElastic


class TestMATS3DElastic(TestMATS3D):
    '''Test the elementary load cases of 3D material models.
    '''

    def _mats_eval_default(self):
        mats_eval = MATS3DElastic(E=34000, nu=0.20)
        return mats_eval

    def test_symmetry_tension(self):
        '''
        Symmetry of the response ... tensile loading.
        '''
        self.assert_symmetry_on_cube_with_clamped_face(load_dirs=[0, 1, 2],
                                                       load=0.001)

    def test_symmetry_bending(self):
        '''
        Symmetry of the response ... out of plane load.
        '''
        self.assert_symmetry_on_cube_with_clamped_face(load_dirs=[1, 2, 0],
                                                       load=0.00016)
        return

    def test_stress_value(self):
        '''Assert a correct value of resulting stress a material point.
        '''
        self.assert_stress_value([340, 0, 0, 0, 0, 0, 0, 0, 0],
                                 n_steps=1, load=0.01)


if __name__ == "__main__":
    import unittest
    #import sys;sys.argv = ['', 'TestMATS3DElastic.test_stress_value']
    unittest.main()
